from io import BytesIO

from breezy import diff, errors, urlutils

try:
    from breezy.transport import NoSuchFile
except ImportError:
    from breezy.errors import NoSuchFile

from breezy.tree import find_previous_path

from .. import util
from ..controllers import TemplatedBranchView


def _process_diff(difftext):
    chunks = []
    chunk = None

    def decode_line(line):
        return line.decode("utf-8", "replace")

    for line in difftext.splitlines():
        if len(line) == 0:
            continue
        if line.startswith(b"+++ ") or line.startswith(b"--- "):
            continue
        if line.startswith(b"@@ "):
            # new chunk
            if chunk is not None:
                chunks.append(chunk)
            chunk = util.Container()
            chunk.diff = []
            split_lines = line.split(b" ")[1:3]
            lines = [int(x.split(b",")[0][1:]) for x in split_lines]
            old_lineno = lines[0]
            new_lineno = lines[1]
        elif line.startswith(b" "):
            chunk.diff.append(
                util.Container(
                    old_lineno=old_lineno,
                    new_lineno=new_lineno,
                    type="context",
                    line=decode_line(line[1:]),
                )
            )
            old_lineno += 1
            new_lineno += 1
        elif line.startswith(b"+"):
            chunk.diff.append(
                util.Container(
                    old_lineno=None,
                    new_lineno=new_lineno,
                    type="insert",
                    line=decode_line(line[1:]),
                )
            )
            new_lineno += 1
        elif line.startswith(b"-"):
            chunk.diff.append(
                util.Container(
                    old_lineno=old_lineno,
                    new_lineno=None,
                    type="delete",
                    line=decode_line(line[1:]),
                )
            )
            old_lineno += 1
        else:
            chunk.diff.append(
                util.Container(
                    old_lineno=None, new_lineno=None, type="unknown", line=repr(line)
                )
            )
    if chunk is not None:
        chunks.append(chunk)
    return chunks


def diff_chunks_for_file(
    repository, filename, compare_revid, revid, context_lines=None
):
    if context_lines is None:
        context_lines = 3
    lines = {}
    compare_tree = repository.revision_tree(compare_revid)
    tree = repository.revision_tree(revid)
    try:
        lines[revid] = tree.get_file_lines(filename)
    except NoSuchFile:
        lines[revid] = []
        lines[compare_revid] = compare_tree.get_file_lines(filename)
    else:
        compare_filename = find_previous_path(tree, compare_tree, filename)
        if compare_filename is not None:
            lines[compare_revid] = compare_tree.get_file_lines(compare_filename)
        else:
            lines[compare_revid] = []

    buffer = BytesIO()
    try:
        diff.internal_diff(
            "",
            lines[compare_revid],
            "",
            lines[revid],
            buffer,
            context_lines=context_lines,
        )
    except errors.BinaryFile:
        difftext = b""
    else:
        difftext = buffer.getvalue()

    return _process_diff(difftext)


class FileDiffUI(TemplatedBranchView):
    template_name = "filediff"
    supports_json = True

    def get_values(self, path, kwargs, headers):
        revid = urlutils.unquote_to_bytes(self.args[0])
        compare_revid = urlutils.unquote_to_bytes(self.args[1])
        filename = urlutils.unquote(self.args[2])

        try:
            context_lines = int(kwargs["context"])
        except (KeyError, ValueError):
            context_lines = None

        chunks = diff_chunks_for_file(
            self._history._branch.repository,
            filename,
            compare_revid,
            revid,
            context_lines=context_lines,
        )

        return {
            "chunks": chunks,
        }



#
# Copyright (C) 2010 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import itertools
from datetime import datetime

from .. import util
from ..controllers.view_ui import ViewUI


class AnnotateUI(ViewUI):
    def annotate_file(self, path, info):
        revid = info["change"].revid
        if not isinstance(revid, bytes):
            raise TypeError(revid)

        tree = self.tree_for(path, revid)

        change_cache = {}
        last_line_revid = None
        last_lineno = None
        message = ""

        revisions = {}

        lineno = 0
        for (line_revid, text), lineno in zip(
            tree.annotate_iter(path), itertools.count(1)
        ):
            if line_revid != last_line_revid:
                last_line_revid = line_revid

                change = change_cache.get(line_revid, None)
                if change is None:
                    changes = self._history.get_changes([line_revid])
                    if changes:
                        change = changes[0]
                    else:
                        change = util.Container(
                            authors="", date=datetime.now(), revno=""
                        )
                    change_cache[line_revid] = change

                try:
                    message = change.comment.splitlines()[0]
                except IndexError:
                    # Comment not present for this revision
                    message = ""

                if last_lineno:
                    # The revspan is of lines between the last revision and this one.
                    # We set the one for the previous revision when we're creating the current revision.
                    revisions[last_lineno].revspan = lineno - last_lineno

                revisions[lineno] = util.Container(change=change, message=message)

                last_lineno = lineno
                last_line_revid = line_revid

        # Zero-size file. Return empty revisions.
        if last_lineno is None:
            return revisions

        # We never set a revspan for the last revision during the loop above, so set it here.
        revisions[last_lineno].revspan = lineno - last_lineno + 1

        return revisions

    def get_values(self, path, kwargs, headers):
        values = super(AnnotateUI, self).get_values(path, kwargs, headers)
        values["annotated"] = self.annotate_file(path, values)

        return values



# Copyright (C) 2008-2011 Canonical Ltd.
#                     (Authored by Martin Albisetti <argentina@gmail.com>)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import time
from io import BytesIO

from breezy.diff import show_diff_trees
from breezy.revision import NULL_REVISION
from paste.request import parse_querystring, path_info_pop

from ..controllers import TemplatedBranchView


class DiffUI(TemplatedBranchView):
    """Class to output a diff for a single file or revisions."""

    def __call__(self, environ, start_response):
        # End of URL is now /diff/<rev_id>?context=<context_lines>
        # or /diff/<rev_id>/<rev_id>?context=<context_lines>
        # This allows users to choose how much context they want to see.
        # Old format was /diff/<rev_id>/<rev_id> or /diff/<rev_id>
        """Default method called from /diff URL."""
        z = time.time()

        args = []
        while True:
            arg = path_info_pop(environ)
            if arg is None:
                break
            args.append(arg)

        numlines = 3  # This is the default.

        opts = parse_querystring(environ)
        for opt in opts:
            if opt[0] == "context":
                try:
                    numlines = int(opt[1])
                except ValueError:
                    pass

        revid_from = args[0]
        # Convert a revno to a revid if we get a revno.
        revid_from = self._history.fix_revid(revid_from)
        change = self._history.get_changes([revid_from])[0]

        if len(args) == 2:
            revid_to = self._history.fix_revid(args[1])
        elif len(change.parents) == 0:
            revid_to = NULL_REVISION
        else:
            revid_to = change.parents[0].revid

        repo = self._branch.branch.repository
        revtree1 = repo.revision_tree(revid_to)
        revtree2 = repo.revision_tree(revid_from)

        diff_content_stream = BytesIO()
        show_diff_trees(
            revtree1,
            revtree2,
            diff_content_stream,
            old_label="",
            new_label="",
            context=numlines,
        )

        content = diff_content_stream.getvalue()

        self.log.info(
            "/diff %r:%r in %r secs with %r context"
            % (revid_from, revid_to, time.time() - z, numlines)
        )

        revno1 = self._history.get_revno(revid_from)
        revno2 = self._history.get_revno(revid_to)
        filename = "%s_%s.diff" % (revno1, revno2)
        headers = [
            ("Content-Type", "application/octet-stream"),
            ("Content-Length", str(len(content))),
            ("Content-Disposition", "attachment; filename=%s" % (filename,)),
        ]
        start_response("200 OK", headers)
        return [content]



# Copyright (C) 2008-2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import json
import tarfile
import tempfile

from paste.fixture import AppError
from paste.httpexceptions import HTTPNotFound
from testtools.matchers import Matcher, Mismatch

from ..apps.branch import BranchWSGIApp
from ..controllers.annotate_ui import AnnotateUI
from .test_simple import BasicTests, TestWithSimpleTree, consume_app


class TestInventoryUI(BasicTests):
    def make_bzrbranch_for_tree_shape(self, shape):
        tree = self.make_branch_and_tree(".")
        self.build_tree(shape)
        tree.smart_add([])
        tree.commit("")
        self.addCleanup(tree.branch.lock_read().unlock)
        return tree.branch

    def make_bzrbranch_and_inventory_ui_for_tree_shape(self, shape, env):
        branch = self.make_bzrbranch_for_tree_shape(shape)
        branch_app = self.make_branch_app(branch)
        return branch, branch_app.lookup_app(env)

    def test_get_filelist(self):
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/files",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        bzrbranch, inv_ui = self.make_bzrbranch_and_inventory_ui_for_tree_shape(
            ["filename"], env
        )
        revtree = bzrbranch.repository.revision_tree(bzrbranch.last_revision())
        self.assertEqual(1, len(inv_ui.get_filelist(revtree, "", "filename", "head")))

    def test_smoke(self):
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/files",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        bzrbranch, inv_ui = self.make_bzrbranch_and_inventory_ui_for_tree_shape(
            ["filename"], env
        )
        start, content = consume_app(inv_ui, env)
        self.assertEqual(("200 OK", [("Content-Type", "text/html")], None), start)
        self.assertContainsRe(content, b"filename")

    def test_no_content_for_HEAD(self):
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/files",
            "REQUEST_METHOD": "HEAD",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        bzrbranch, inv_ui = self.make_bzrbranch_and_inventory_ui_for_tree_shape(
            ["filename"], env
        )
        start, content = consume_app(inv_ui, env)
        self.assertEqual(("200 OK", [("Content-Type", "text/html")], None), start)
        self.assertEqual(b"", content)

    def test_get_values_smoke(self):
        branch = self.make_bzrbranch_for_tree_shape(["a-file"])
        branch_app = self.make_branch_app(branch)
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/files",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        inv_ui = branch_app.lookup_app(env)
        inv_ui.parse_args(env)
        values = inv_ui.get_values("", {}, {})
        self.assertEqual("a-file", values["filelist"][0].filename)

    def test_json_render_smoke(self):
        branch = self.make_bzrbranch_for_tree_shape(["a-file"])
        branch_app = self.make_branch_app(branch)
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/+json/files",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        inv_ui = branch_app.lookup_app(env)
        self.assertOkJsonResponse(inv_ui, env)


class TestRevisionUI(BasicTests):
    def make_branch_app_for_revision_ui(self, shape1, shape2):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents(shape1)
        tree.smart_add([])
        tree.commit("msg 1", rev_id=b"rev-1")
        self.build_tree_contents(shape2)
        tree.smart_add([])
        tree.commit("msg 2", rev_id=b"rev-2")
        branch = tree.branch
        self.addCleanup(branch.lock_read().unlock)
        return self.make_branch_app(branch)

    def test_get_values(self):
        branch_app = self.make_branch_app_for_revision_ui([], [])
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/revision/2",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        rev_ui = branch_app.lookup_app(env)
        rev_ui.parse_args(env)
        self.assertIsInstance(rev_ui.get_values("", {}, []), dict)

    def test_add_template_values(self):
        branch_app = self.make_branch_app_for_revision_ui(
            [("file", b"content\n")], [("file", b"new content\n")]
        )
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/revision/1/non-existent-file",
            "QUERY_STRING": "start_revid=1",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revision_ui = branch_app.lookup_app(env)
        path = revision_ui.parse_args(env)
        values = revision_ui.get_values(path, revision_ui.kwargs, {})
        revision_ui.add_template_values(values)
        self.assertIsNone(values["diff_chunks"])

    def test_add_template_values_with_changes(self):
        branch_app = self.make_branch_app_for_revision_ui(
            [("file", b"content\n")], [("file", b"new content\n")]
        )
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/revision/1/file",
            "QUERY_STRING": "start_revid=1",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revision_ui = branch_app.lookup_app(env)
        path = revision_ui.parse_args(env)
        values = revision_ui.get_values(path, revision_ui.kwargs, {})
        revision_ui.add_template_values(values)
        self.assertEqual(len(values["diff_chunks"]), 1)

    def test_add_template_values_with_non_ascii(self):
        branch_app = self.make_branch_app_for_revision_ui(
            [("skr\xe1", b"content\n")], [("skr\xe1", b"new content\n")]
        )
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/revision/1",
            "QUERY_STRING": "start_revid=1",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revision_ui = branch_app.lookup_app(env)
        path = revision_ui.parse_args(env)
        values = revision_ui.get_values(path, revision_ui.kwargs, {})
        revision_ui.add_template_values(values)
        self.assertEqual(
            json.loads(values["link_data"]),
            {
                "diff-0": "rev-1/null%253A/%252F",
                "diff-1": "rev-1/null%253A/skr%25C3%25A1",
            },
        )
        self.assertEqual(
            json.loads(values["path_to_id"]), {"/": "diff-0", "skr\xe1": "diff-1"}
        )

    def test_get_values_smoke(self):
        branch_app = self.make_branch_app_for_revision_ui(
            [("file", b"content\n"), ("other-file", b"other\n")],
            [("file", b"new content\n")],
        )
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/revision/head:",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revision_ui = branch_app.lookup_app(env)
        revision_ui.parse_args(env)
        values = revision_ui.get_values("", {}, {})

        self.assertEqual(values["revid"], "rev-2")
        self.assertEqual(values["change"].comment, "msg 2")
        self.assertEqual(values["file_changes"].modified[0].filename, "file")
        self.assertEqual(values["merged_in"], None)

    def test_json_render_smoke(self):
        branch_app = self.make_branch_app_for_revision_ui(
            [("file", b"content\n"), ("other-file", b"other\n")],
            [("file", b"new content\n")],
        )
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/+json/revision/head:",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revision_ui = branch_app.lookup_app(env)
        self.assertOkJsonResponse(revision_ui, env)


class TestAnnotateUI(BasicTests):
    def make_annotate_ui_for_file_history(self, filename, rev_ids_texts):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(filename, "")])
        tree.add([filename])
        for rev_id, text, message in rev_ids_texts:
            self.build_tree_contents([(filename, text)])
            tree.commit(rev_id=rev_id, message=message)
        tree.branch.lock_read()
        self.addCleanup(tree.branch.unlock)
        branch_app = BranchWSGIApp(tree.branch, friendly_name="test_name")
        return AnnotateUI(branch_app, branch_app.get_history)

    def test_annotate_file(self):
        history = [(b"rev1", b"old\nold\n", "."), (b"rev2", b"new\nold\n", ".")]
        ann_ui = self.make_annotate_ui_for_file_history("filename", history)
        # A lot of this state is set up by __call__, but we'll do it directly
        # here.
        ann_ui.args = ["rev2"]
        annotate_info = ann_ui.get_values("filename", kwargs={}, headers={})
        annotated = annotate_info["annotated"]
        self.assertEqual(2, len(annotated))
        self.assertEqual("2", annotated[1].change.revno)
        self.assertEqual("1", annotated[2].change.revno)

    def test_annotate_empty_comment(self):
        # Testing empty comment handling without breaking
        history = [(b"rev1", b"old\nold\n", "."), (b"rev2", b"new\nold\n", "")]
        ann_ui = self.make_annotate_ui_for_file_history("filename", history)
        ann_ui.args = ["rev2"]
        ann_ui.get_values("filename", kwargs={}, headers={})

    def test_annotate_file_zero_sized(self):
        # Test against a zero-sized file without breaking. No annotation
        # must be present.
        history = [(b"rev1", b"", ".")]
        ann_ui = self.make_annotate_ui_for_file_history("filename", history)
        ann_ui.args = ["rev1"]
        annotate_info = ann_ui.get_values("filename", kwargs={}, headers={})
        annotated = annotate_info["annotated"]
        self.assertEqual(0, len(annotated))

    def test_annotate_nonexistent_file(self):
        history = [(b"rev1", b"", ".")]
        ann_ui = self.make_annotate_ui_for_file_history("filename", history)
        ann_ui.args = ["rev1"]
        self.assertRaises(HTTPNotFound, ann_ui.get_values, "not-filename", {}, {})

    def test_annotate_nonexistent_rev(self):
        history = [(b"rev1", b"", ".")]
        ann_ui = self.make_annotate_ui_for_file_history("filename", history)
        ann_ui.args = ["norev"]
        self.assertRaises(HTTPNotFound, ann_ui.get_values, "not-filename", {}, {})


class TestFileDiffUI(BasicTests):
    def make_branch_app_for_filediff_ui(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        rev1 = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("filename", None, "file", b"content\n")),
            ],
            message="First commit.",
        )
        rev2 = builder.build_snapshot(
            None, [("modify", ("filename", b"new content\n"))]
        )
        builder.finish_series()
        branch = builder.get_branch()
        self.addCleanup(branch.lock_read().unlock)
        return self.make_branch_app(branch), (rev1, rev2)

    def test_get_values_smoke(self):
        branch_app, (rev1, rev2) = self.make_branch_app_for_filediff_ui()
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/+filediff/{}/{}/filename".format(
                rev2.decode("utf-8"), rev1.decode("utf-8")
            ),
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        filediff_ui = branch_app.lookup_app(env)
        filediff_ui.parse_args(env)
        values = filediff_ui.get_values("", {}, {})
        chunks = values["chunks"]
        self.assertEqual("insert", chunks[0].diff[1].type)
        self.assertEqual("new content", chunks[0].diff[1].line)

    def test_json_render_smoke(self):
        branch_app, (rev1, rev2) = self.make_branch_app_for_filediff_ui()
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/+json/+filediff/{}/{}/filename".format(
                rev2.decode("utf-8"), rev1.decode("utf-8")
            ),
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        filediff_ui = branch_app.lookup_app(env)
        self.assertOkJsonResponse(filediff_ui, env)


class TestRevLogUI(BasicTests):
    def make_branch_app_for_revlog_ui(self):
        builder = self.make_branch_builder("branch")
        builder.start_series()
        revid = builder.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("filename", None, "file", b"content\n")),
            ],
            message="First commit.",
        )
        builder.finish_series()
        branch = builder.get_branch()
        self.addCleanup(branch.lock_read().unlock)
        return self.make_branch_app(branch), revid

    def test_get_values_smoke(self):
        branch_app, revid = self.make_branch_app_for_revlog_ui()
        env = {
            "SCRIPT_NAME": "/",
            "PATH_INFO": "/+revlog/%s" % revid.decode("utf-8"),
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revlog_ui = branch_app.lookup_app(env)
        revlog_ui.parse_args(env)
        values = revlog_ui.get_values("", {}, {})
        self.assertEqual(values["file_changes"].added[1].filename, "filename")
        self.assertEqual(values["entry"].comment, "First commit.")

    def test_json_render_smoke(self):
        branch_app, revid = self.make_branch_app_for_revlog_ui()
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/+json/+revlog/%s" % revid.decode("utf-8"),
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }
        revlog_ui = branch_app.lookup_app(env)
        self.assertOkJsonResponse(revlog_ui, env)


class TestControllerHooks(BasicTests):
    def test_dummy_hook(self):
        return
        # A hook that returns None doesn't influence the searching for
        # a controller.
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/custom",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }

        def myhook(app, environ):
            return None

        branch = self.make_branch(".")
        self.addCleanup(branch.lock_read().unlock)
        app = self.make_branch_app(branch)
        self.addCleanup(
            BranchWSGIApp.hooks.uninstall_named_hook, "controller", "captain hook"
        )
        BranchWSGIApp.hooks.install_named_hook("controller", myhook, "captain hook")
        self.assertRaises(KeyError, app.lookup_app, env)

    def test_working_hook(self):
        # A hook can provide an app to use for a particular request.
        env = {
            "SCRIPT_NAME": "",
            "PATH_INFO": "/custom",
            "REQUEST_METHOD": "GET",
            "wsgi.url_scheme": "http",
            "SERVER_NAME": "localhost",
            "SERVER_PORT": "80",
        }

        def myhook(app, environ):
            return "I am hooked"

        branch = self.make_branch(".")
        self.addCleanup(branch.lock_read().unlock)
        app = self.make_branch_app(branch)
        self.addCleanup(
            BranchWSGIApp.hooks.uninstall_named_hook, "controller", "captain hook"
        )
        BranchWSGIApp.hooks.install_named_hook("controller", myhook, "captain hook")
        self.assertEqual("I am hooked", app.lookup_app(env))


class MatchesDownloadHeaders(Matcher):
    def __init__(self, expect_filename, expect_mimetype):
        self.expect_filename = expect_filename
        self.expect_mimetype = expect_mimetype

    def match(self, response):
        # Maybe the c-t should be more specific, but this is probably good for
        # making sure it gets saved without the client trying to decompress it
        # or anything.
        if (
            response.header("Content-Type") == self.expect_mimetype
            and response.header("Content-Disposition")
            == "attachment; filename*=utf-8''" + self.expect_filename
        ):
            pass
        else:
            return Mismatch("wrong response headers: %r" % response.headers)

    def __str__(self):
        return "MatchesDownloadHeaders({!r}, {!r})".format(
            self.expect_filename, self.expect_mimetype
        )


class TestDownloadUI(TestWithSimpleTree):
    def test_download(self):
        app = self.setUpLoggerhead()
        response = app.get("/download/1/myfilename")
        self.assertEqual(
            b"some\nmultiline\ndata\nwith<htmlspecialchars\n", response.body
        )
        self.assertThat(
            response, MatchesDownloadHeaders("myfilename", "application/octet-stream")
        )

    def test_download_bad_revision(self):
        app = self.setUpLoggerhead()
        e = self.assertRaises(AppError, app.get, "/download/norev/myfilename")
        self.assertContainsRe(str(e), "404 Not Found")

    def test_download_bad_filename(self):
        app = self.setUpLoggerhead()
        e = self.assertRaises(AppError, app.get, "/download/1/notmyfilename")
        self.assertContainsRe(str(e), "404 Not Found")

    def test_download_from_subdirectory(self):
        app = self.setUpLoggerhead()
        response = app.get("/download/1/folder/myfilename")
        self.assertEqual(
            b"some\nmultiline\ndata\nwith<htmlspecialchars\n", response.body
        )
        self.assertThat(
            response, MatchesDownloadHeaders("myfilename", "application/octet-stream")
        )


class IsTarfile(Matcher):
    def __init__(self, compression):
        self.compression = compression

    def match(self, content_bytes):
        f = tempfile.NamedTemporaryFile()
        try:
            f.write(content_bytes)
            f.flush()
            tarfile.open(f.name, mode="r|" + self.compression)
        finally:
            f.close()


class TestDownloadTarballUI(TestWithSimpleTree):
    def test_download_tarball(self):
        # Tarball downloads are enabled by default.
        app = self.setUpLoggerhead()
        response = app.get("/tarball")
        self.assertThat(response.body, IsTarfile("gz"))
        self.assertThat(
            response, MatchesDownloadHeaders("branch.tgz", "application/octet-stream")
        )

    def test_download_tarball_of_version(self):
        app = self.setUpLoggerhead()
        response = app.get("/tarball/1")
        self.assertThat(response.body, IsTarfile("gz"))
        self.assertThat(
            response,
            MatchesDownloadHeaders("branch-r1.tgz", "application/octet-stream"),
        )

    def test_download_tarball_forbidden(self):
        app = self.setUpLoggerhead(export_tarballs=False)
        e = self.assertRaises(AppError, app.get, "/tarball")
        self.assertContainsRe(
            str(e), "(?s).*403 Forbidden.*Tarball downloads are not allowed"
        )



from breezy import osutils, urlutils

from ..controllers import TemplatedBranchView


class RevLogUI(TemplatedBranchView):
    template_name = "revlog"
    supports_json = True

    def get_values(self, path, kwargs, headers):
        history = self._history

        revid = urlutils.unquote_to_bytes(osutils.safe_utf8(self.args[0]))

        change = history.get_changes([revid])[0]
        file_changes = history.get_file_changes(change)
        history.add_branch_nicks(change)

        return {
            "entry": change,
            "file_changes": file_changes,
            "revid": revid,
        }



#
# Copyright (C) 2008  Canonical Ltd.
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import posixpath

from breezy import errors, osutils, urlutils
from breezy.revision import is_null as is_null_rev
from paste.httpexceptions import HTTPMovedPermanently, HTTPNotFound

from .. import util
from ..controllers import TemplatedBranchView


def dirname(path):
    if path is not None:
        path = path.rstrip("/")
        path = urlutils.escape(posixpath.dirname(path))
    return path


class InventoryUI(TemplatedBranchView):
    template_name = "inventory"
    supports_json = True

    def get_filelist(self, tree, path, sort_type, revno_url):
        """
        return the list of all files (and their attributes) within a given
        path subtree.

        @param tree: The tree
        @param path: The path of a directory within the tree.
        @param sort_type: How to sort the results... XXX.
        """
        file_list = []

        if tree.kind(path) != "directory":
            raise HTTPMovedPermanently(
                self._branch.context_url(["/view", revno_url, path])
            )

        revid_set = set()

        child_entries = []

        for entry in tree.iter_child_entries(path):
            child_path = osutils.pathjoin(path, entry.name)
            child_revision = tree.get_file_revision(child_path)
            revid_set.add(child_revision)
            child_entries.append((child_path, entry, child_revision))

        change_dict = {}
        for change in self._history.get_changes(list(revid_set)):
            change_dict[change.revid] = change

        for child_path, entry, child_revision in child_entries:
            pathname = entry.name
            contents_changed_rev = None
            if entry.kind == "directory":
                pathname += "/"
                size = None
            else:
                size = entry.text_size

            change_dict[child_revision].timestamp

            # TODO: For the JSON rendering, this inlines the "change" aka
            # revision information attached to each file. Consider either
            # pulling this out as a separate changes dict, or possibly just
            # including the revision id and having a separate request to get
            # back the revision info.
            file = util.Container(
                filename=entry.name,
                executable=entry.executable,
                kind=entry.kind,
                absolutepath=child_path,
                size=size,
                revid=child_revision,
                change=change_dict[child_revision],
                contents_changed_rev=contents_changed_rev,
            )
            file_list.append(file)

        if sort_type == "filename":
            file_list.sort(key=lambda x: x.filename.lower())  # case-insensitive
        elif sort_type == "size":

            def size_key(x):
                if x.size is None:
                    return -1
                return x.size

            file_list.sort(key=size_key)
        elif sort_type == "date":
            file_list.sort(key=lambda x: x.change.date, reverse=True)

        if sort_type != "date":
            # Don't always sort directories first.
            file_list.sort(key=lambda x: x.kind != "directory")

        return file_list

    def get_values(self, path, kwargs, headers):
        history = self._history
        branch = history._branch
        try:
            revid = self.get_revid()
            rev_tree = branch.repository.revision_tree(revid)
        except errors.NoSuchRevision:
            raise HTTPNotFound()

        start_revid = kwargs.get("start_revid", None)
        sort_type = kwargs.get("sort", "filename")

        if path is None:
            path = "/"

        path = path.rstrip("/")
        if not rev_tree.has_filename(path) and not is_null_rev(revid):
            raise HTTPNotFound()

        # Are we at the top of the tree
        if path in ["/", ""]:
            updir = None
        else:
            updir = dirname(path)

        if not is_null_rev(revid):
            change = history.get_changes([revid])[0]
            # If we're looking at the tip, use head: in the URL instead
            if revid == branch.last_revision():
                revno_url = "head:"
            else:
                revno_url = history.get_revno(revid)
            history.add_branch_nicks(change)
            filelist = self.get_filelist(rev_tree, path, sort_type, revno_url)

        else:
            start_revid = None
            change = None
            path = "/"
            updir = None
            revno_url = "head:"
            filelist = []

        return {
            "revid": revid,
            "revno_url": revno_url,
            "change": change,
            "path": path,
            "updir": updir,
            "filelist": filelist,
            "start_revid": start_revid,
        }

    def add_template_values(self, values):
        super(InventoryUI, self).add_template_values(values)
        # Directory Breadcrumbs
        directory_breadcrumbs = util.directory_breadcrumbs(
            self._branch.friendly_name, self._branch.is_root, "files"
        )

        path = values["path"]
        revid = values["revid"]
        # no navbar for revisions
        navigation = util.Container()

        if is_null_rev(revid):
            branch_breadcrumbs = []
        else:
            # Create breadcrumb trail for the path within the branch
            branch = self._history._branch
            rev_tree = branch.repository.revision_tree(revid)
            branch_breadcrumbs = util.branch_breadcrumbs(path, rev_tree, "files")
        values.update(
            {
                "fileview_active": True,
                "directory_breadcrumbs": directory_breadcrumbs,
                "branch_breadcrumbs": branch_breadcrumbs,
                "navigation": navigation,
            }
        )



# Copyright (C) 2008-2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#
"""Serve branches at urls that mimic a transport's file system layout."""

import threading

from breezy import branch, errors, lru_cache, urlutils
from breezy.config import LocationConfig
from breezy.transport import get_transport
from breezy.transport.http import wsgi
from paste import httpexceptions, urlparser
from paste.request import path_info_pop

from .. import util
from ..apps import favicon_app, robots_app, static_app
from ..apps.branch import BranchWSGIApp
from ..controllers.directory_ui import DirectoryUI


class BranchesFromTransportServer(object):
    def __init__(self, transport, root, name=None):
        self.transport = transport
        self.root = root
        self.name = name
        self._config = root._config

    def app_for_branch(self, branch):
        if not self.name:
            name = branch._get_nick(local=True)
            is_root = True
        else:
            name = self.name
            is_root = False
        branch_app = BranchWSGIApp(
            branch,
            name,
            {"cachepath": self._config.SQL_DIR},
            self.root.graph_cache,
            is_root=is_root,
            use_cdn=self._config.get_option("use_cdn"),
        )
        return branch_app.app

    def app_for_non_branch(self, environ):
        segment = path_info_pop(environ)
        if segment is None:
            raise httpexceptions.HTTPMovedPermanently.relative_redirect(
                environ["SCRIPT_NAME"] + "/", environ
            )
        elif segment == "":
            if self.name:
                name = self.name
            else:
                name = "/"
            return DirectoryUI(environ["loggerhead.static.url"], self.transport, name)
        else:
            new_transport = self.transport.clone(segment)
            if self.name:
                new_name = urlutils.join(self.name, segment)
            else:
                new_name = "/" + segment
            return BranchesFromTransportServer(new_transport, self.root, new_name)

    def app_for_bazaar_data(self, relpath):
        if relpath == "/.bzr/smart":
            root_transport = get_transport_for_thread(self.root.base)
            wsgi_app = wsgi.SmartWSGIApp(root_transport)
            return wsgi.RelpathSetter(wsgi_app, "", "loggerhead.path_info")
        else:
            # TODO: Use something here that uses the transport API
            # rather than relying on the local filesystem API.
            base = self.transport.base
            try:
                path = util.local_path_from_url(base)
            except errors.InvalidURL:
                raise httpexceptions.HTTPNotFound()
            else:
                return urlparser.make_static(None, path)

    def check_serveable(self, config):
        if not config.get_user_option_as_bool("http_serve", default=True):
            raise httpexceptions.HTTPNotFound()

    def __call__(self, environ, start_response):
        path = environ["PATH_INFO"]
        try:
            b = branch.Branch.open_from_transport(self.transport)
        except errors.NotBranchError:
            if path.startswith("/.bzr"):
                self.check_serveable(LocationConfig(self.transport.base))
                return self.app_for_bazaar_data(path)(environ, start_response)
            if not self.transport.listable() or not self.transport.has("."):
                raise httpexceptions.HTTPNotFound()
            return self.app_for_non_branch(environ)(environ, start_response)
        else:
            self.check_serveable(b.get_config())
            if path.startswith("/.bzr"):
                return self.app_for_bazaar_data(path)(environ, start_response)
            else:
                return self.app_for_branch(b)(environ, start_response)


_transport_store = threading.local()


def get_transport_for_thread(base):
    thread_transports = getattr(_transport_store, "transports", None)
    if thread_transports is None:
        thread_transports = _transport_store.transports = {}
    if base in thread_transports:
        return thread_transports[base]
    transport = get_transport(base)
    thread_transports[base] = transport
    return transport


class BranchesFromTransportRoot(object):
    def __init__(self, base, config):
        self.graph_cache = lru_cache.LRUCache(10)
        self.base = base
        self._config = config

    def __call__(self, environ, start_response):
        environ["loggerhead.static.url"] = environ["SCRIPT_NAME"]
        environ["loggerhead.path_info"] = environ["PATH_INFO"]
        if environ["PATH_INFO"].startswith("/static/"):
            segment = path_info_pop(environ)
            assert segment == "static"
            return static_app(environ, start_response)
        elif environ["PATH_INFO"] == "/favicon.ico":
            return favicon_app(environ, start_response)
        elif environ["PATH_INFO"] == "/robots.txt":
            return robots_app(environ, start_response)
        else:
            transport = get_transport_for_thread(self.base)
            return BranchesFromTransportServer(transport, self)(environ, start_response)


class UserBranchesFromTransportRoot(object):
    def __init__(self, base, config):
        self.graph_cache = lru_cache.LRUCache(10)
        self.base = base
        self._config = config
        self.trunk_dir = config.get_option("trunk_dir")

    def __call__(self, environ, start_response):
        environ["loggerhead.static.url"] = environ["SCRIPT_NAME"]
        environ["loggerhead.path_info"] = environ["PATH_INFO"]
        path_info = environ["PATH_INFO"]
        if path_info.startswith("/static/"):
            segment = path_info_pop(environ)
            assert segment == "static"
            return static_app(environ, start_response)
        elif path_info == "/favicon.ico":
            return favicon_app(environ, start_response)
        elif environ["PATH_INFO"] == "/robots.txt":
            return robots_app(environ, start_response)
        else:
            transport = get_transport_for_thread(self.base)
            # segments starting with ~ are user branches
            if path_info.startswith("/~"):
                segment = path_info_pop(environ)
                return BranchesFromTransportServer(
                    transport.clone(segment[1:]), self, segment
                )(environ, start_response)
            else:
                return BranchesFromTransportServer(
                    transport.clone(self.trunk_dir), self
                )(environ, start_response)



from ..zptsupport import load_template

RENDERED = "<html>\n<head>\n<title>%s</title>\n</head>\n\
<body>\n<div>Hello, %s</div>\n</body>\n</html>"


def test_template_lookup():
    template = load_template("loggerhead.tests.simple")
    assert template
    TITLE = "test"
    NAME = "World"
    info = dict(title=TITLE, name=NAME)
    s = template.expand(**info)
    assert s.startswith(RENDERED % (TITLE, NAME))



#
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import os

from breezy.errors import BinaryFile

try:
    from breezy.transport import NoSuchFile
except ImportError:
    from breezy.errors import NoSuchFile

import breezy.textfile
from breezy import osutils
from paste.httpexceptions import HTTPBadRequest, HTTPMovedPermanently, HTTPNotFound

from ..controllers import TemplatedBranchView

try:
    from ..highlight import highlight
except ImportError:
    highlight = None
from .. import util


class ViewUI(TemplatedBranchView):
    template_name = "view"

    def tree_for(self, path, revid):
        if not isinstance(path, str):
            raise TypeError(path)
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        return self._history._branch.repository.revision_tree(revid)

    def text_lines(self, path, revid):
        file_name = os.path.basename(path)

        tree = self.tree_for(path, revid)
        file_text = tree.get_file_text(path)

        encoding = "utf-8"
        try:
            file_text.decode(encoding)
        except UnicodeDecodeError:
            encoding = "iso-8859-15"
            file_text.decode(encoding)

        file_lines = osutils.split_lines(file_text)
        # This can throw breezy.errors.BinaryFile (which our caller catches).
        breezy.textfile.check_text_lines(file_lines)

        file_text = file_text.decode(encoding)
        file_lines = osutils.split_lines(file_text)

        if highlight is not None:
            hl_lines = highlight(file_name, file_text, encoding)
            # highlight strips off extra newlines at the end of the file.
            extra_lines = len(file_lines) - len(hl_lines)
            hl_lines.extend([""] * extra_lines)
        else:
            hl_lines = [util.html_escape(line) for line in file_lines]

        return hl_lines

    def file_contents(self, path, revid):
        try:
            file_lines = self.text_lines(path, revid)
        except BinaryFile:
            # bail out; this isn't displayable text
            return ["(This is a binary file.)"]

        return file_lines

    def get_values(self, path, kwargs, headers):
        history = self._history
        branch = history._branch
        revid = self.get_revid()
        if path is None:
            raise HTTPBadRequest("No filename provided to view")

        if not history.file_exists(revid, path):
            raise HTTPNotFound()

        filename = os.path.basename(path)

        change = history.get_changes([revid])[0]
        # If we're looking at the tip, use head: in the URL instead
        if revid == branch.last_revision():
            revno_url = "head:"
        else:
            revno_url = history.get_revno(revid)

        # Directory Breadcrumbs
        directory_breadcrumbs = util.directory_breadcrumbs(
            self._branch.friendly_name, self._branch.is_root, "files"
        )

        tree = history.revision_tree(revid)

        # Create breadcrumb trail for the path within the branch
        branch_breadcrumbs = util.branch_breadcrumbs(path, tree, "files")

        try:
            if tree.kind(path) == "directory":
                raise HTTPMovedPermanently(
                    self._branch.context_url(["/files", revno_url, path])
                )
        except NoSuchFile:
            raise HTTPNotFound()

        # no navbar for revisions
        navigation = util.Container()

        return {
            # In AnnotateUI, "annotated" is a dictionary mapping lines to
            # changes.  We exploit the fact that bool({}) is False when
            # checking whether we're in "annotated" mode.
            "annotated": {},
            "revno_url": revno_url,
            "file_path": path,
            "filename": filename,
            "navigation": navigation,
            "change": change,
            "contents": self.file_contents(path, revid),
            "fileview_active": True,
            "directory_breadcrumbs": directory_breadcrumbs,
            "branch_breadcrumbs": branch_breadcrumbs,
        }



#
# Copyright (C) 2008  Canonical Ltd.
#                     (Authored by Martin Albisetti <argentina@gmail.com)
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

from __future__ import print_function

import base64
import datetime
import logging
import os
import re
import struct
import subprocess
import sys
import threading
import time
from xml.etree import ElementTree as ET

import bleach
from breezy import urlutils

log = logging.getLogger("loggerhead.controllers")


def fix_year(year):
    if year < 70:
        year += 2000
    if year < 100:
        year += 1900
    return year


# Display of times.

# date_day -- just the day
# date_time -- full date with time (UTC)
#
# approximatedate -- for use in tables
#
# approximatedate return an elementtree <span> Element
# with the full date (UTC) in a tooltip.


def date_day(value):
    return value.strftime("%Y-%m-%d")


def date_time(value):
    if value is not None:
        # Note: this assumes that the value is UTC in some fashion.
        return value.strftime("%Y-%m-%d %H:%M:%S UTC")
    else:
        return "N/A"


def _approximatedate(date):
    if date is None:
        return "Never"
    delta = datetime.datetime.utcnow() - date
    future = delta < datetime.timedelta(0, 0, 0)
    delta = abs(delta)
    years = delta.days // 365
    months = delta.days // 30  # This is approximate.
    days = delta.days
    hours = delta.seconds // 3600
    minutes = (delta.seconds - (3600 * hours)) / 60
    seconds = delta.seconds % 60
    result = ""
    if future:
        result += "in "
    if years != 0:
        amount = years
        unit = "year"
    elif months != 0:
        amount = months
        unit = "month"
    elif days != 0:
        amount = days
        unit = "day"
    elif hours != 0:
        amount = hours
        unit = "hour"
    elif minutes != 0:
        amount = minutes
        unit = "minute"
    else:
        amount = seconds
        unit = "second"
    if amount != 1:
        unit += "s"
    result += "%s %s" % (int(amount), unit)
    if not future:
        result += " ago"
    return result


def _wrap_with_date_time_title(date, formatted_date):
    elem = ET.Element("span")
    elem.text = formatted_date
    elem.set("title", date_time(date))
    return elem


def approximatedate(date):
    # FIXME: Returns an object instead of a string
    return _wrap_with_date_time_title(date, _approximatedate(date))


class Container(object):
    """
    Convert a dict into an object with attributes.
    """

    def __init__(self, _dict=None, **kw):
        self._properties = {}
        if _dict is not None:
            for key, value in _dict.items():
                setattr(self, key, value)
        for key, value in kw.items():
            setattr(self, key, value)

    def __repr__(self):
        out = "{ "
        for key, value in self.__dict__.items():
            if key.startswith("_") or (
                getattr(self.__dict__[key], "__call__", None) is not None
            ):
                continue
            out += "%r => %r, " % (key, value)
        out += "}"
        return out

    def __getattr__(self, attr):
        """Used for handling things that aren't already available."""
        if attr.startswith("_") or attr not in self._properties:
            raise AttributeError("No attribute: %s" % (attr,))
        val = self._properties[attr](self, attr)
        setattr(self, attr, val)
        return val

    def _set_property(self, attr, prop_func):
        """Set a function that will be called when an attribute is desired.

        We will cache the return value, so the function call should be
        idempotent. We will pass 'self' and the 'attr' name when triggered.
        """
        if attr.startswith("_"):
            raise ValueError("Cannot create properties that start with _")
        self._properties[attr] = prop_func


def trunc(text, limit=10):
    if len(text) <= limit:
        return text
    return text[:limit] + "..."


STANDARD_PATTERN = re.compile(r"^(.*?)\s*<(.*?)>\s*$")
EMAIL_PATTERN = re.compile(r"[-\w\d\+_!%\.]+@[-\w\d\+_!%\.]+")


def hide_email(email):
    """
    try to obscure any email address in a bazaar committer's name.
    """
    m = STANDARD_PATTERN.search(email)
    if m is not None:
        name = m.group(1)
        email = m.group(2)
        return name
    m = EMAIL_PATTERN.search(email)
    if m is None:
        # can't find an email address in here
        return email
    username, domain = m.group(0).split("@")
    domains = domain.split(".")
    if len(domains) >= 2:
        return "%s at %s" % (username, domains[-2])
    return "%s at %s" % (username, domains[0])


def hide_emails(emails):
    """
    try to obscure any email address in a list of bazaar committers' names.
    """
    result = []
    for email in emails:
        result.append(hide_email(email))
    return result


# only do this if unicode turns out to be a problem
# _BADCHARS_RE = re.compile(ur'[\u007f-\uffff]')

# Can't be a dict; &amp; needs to be done first.
html_entity_subs = [
    ("&", "&amp;"),
    ('"', "&quot;"),
    ("'", "&#39;"),  # &apos; is defined in XML, but not HTML.
    (">", "&gt;"),
    ("<", "&lt;"),
]


def html_escape(s):
    """Transform dangerous (X)HTML characters into entities.

    Like cgi.escape, except also escaping \" and '. This makes it safe to use
    in both attribute and element content.

    If you want to safely fill a format string with escaped values, use
    html_format instead
    """
    for char, repl in html_entity_subs:
        s = s.replace(char, repl)
    return s


def html_format(template, *args):
    """Safely format an HTML template string, escaping the arguments.

    The template string must not be user-controlled; it will not be escaped.
    """
    return template % tuple(html_escape(arg) for arg in args)


# FIXME: get rid of this method; use fixed_width() and avoid XML().


def html_clean(s):
    """
    clean up a string for html display.  expand any tabs, encode any html
    entities, and replace spaces with '&nbsp;'.  this is primarily for use
    in displaying monospace text.
    """
    s = html_escape(s.expandtabs())
    s = s.replace(" ", "&nbsp;")
    return s


NONBREAKING_SPACE = "\N{NO-BREAK SPACE}"


def fill_div(s):
    """
    CSS is stupid. In some cases we need to replace an empty value with
    a non breaking space (&nbsp;). There has to be a better way of doing this.

    return: the same value received if not empty, and a '&nbsp;' if it is.
    """
    if s is None:
        return "&nbsp;"
    elif isinstance(s, int):
        return s
    elif not s.strip():
        return "&nbsp;"
    elif isinstance(s, bytes):
        try:
            s = s.decode("utf-8")
        except UnicodeDecodeError:
            s = s.decode("iso-8859-15")
        return s
    elif isinstance(s, str):
        return s
    else:
        return repr(s)


def fixed_width(s):
    """
    expand tabs and turn spaces into "non-breaking spaces", so browsers won't
    chop up the string.
    """
    if not isinstance(s, str):
        # this kinda sucks.  file contents are just binary data, and no
        # encoding metadata is stored, so we need to guess.  this is probably
        # okay for most code, but for people using things like KOI-8, this
        # will display gibberish.  we have no way of detecting the correct
        # encoding to use.
        try:
            s = s.decode("utf-8")
        except UnicodeDecodeError:
            s = s.decode("iso-8859-15")

    s = html_escape(s).expandtabs().replace(" ", NONBREAKING_SPACE)

    return bleach.clean(s).replace("\n", "<br/>")


def fake_permissions(kind, executable):
    # fake up unix-style permissions given only a "kind" and executable bit
    if kind == "directory":
        return "drwxr-xr-x"
    if executable:
        return "-rwxr-xr-x"
    return "-rw-r--r--"


def b64(s):
    s = base64.encodestring(s).replace("\n", "")
    while (len(s) > 0) and (s[-1] == "="):
        s = s[:-1]
    return s


def uniq(uniqs, s):
    """
    turn a potentially long string into a unique smaller string.
    """
    if s in uniqs:
        return uniqs[s]
    uniqs[type(None)] = next = uniqs.get(type(None), 0) + 1
    x = struct.pack(">I", next)
    while (len(x) > 1) and (x[0] == "\x00"):
        x = x[1:]
    uniqs[s] = b64(x)
    return uniqs[s]


KILO = 1024
MEG = 1024 * KILO
GIG = 1024 * MEG
P95_MEG = int(0.9 * MEG)
P95_GIG = int(0.9 * GIG)


def human_size(size, min_divisor=0):
    size = int(size)
    if (size == 0) and (min_divisor == 0):
        return "Empty"
    if (size < 1024) and (min_divisor == 0):
        return str(size) + " bytes"

    if (size >= P95_GIG) or (min_divisor >= GIG):
        divisor = GIG
    elif (size >= P95_MEG) or (min_divisor >= MEG):
        divisor = MEG
    else:
        divisor = KILO

    dot = size % divisor
    base = size - dot
    dot = dot * 10 // divisor
    base //= divisor
    if dot >= 10:
        base += 1
        dot -= 10

    out = str(base)
    if (base < 100) and (dot != 0):
        out += ".%d" % (dot,)
    if divisor == KILO:
        out += " KB"
    elif divisor == MEG:
        out += " MB"
    elif divisor == GIG:
        out += " GB"
    return out


def local_path_from_url(url):
    """Convert Bazaar URL to local path, ignoring readonly+ prefix"""
    readonly_prefix = "readonly+"
    if url.startswith(readonly_prefix):
        url = url[len(readonly_prefix) :]
    return urlutils.local_path_from_url(url)


def fill_in_navigation(navigation):
    """
    given a navigation block (used by the template for the page header), fill
    in useful calculated values.
    """
    if navigation.revid in navigation.revid_list:  # XXX is this always true?
        navigation.position = navigation.revid_list.index(navigation.revid)
    else:
        navigation.position = 0
    navigation.count = len(navigation.revid_list)
    navigation.page_position = navigation.position // navigation.pagesize + 1
    navigation.page_count = (
        len(navigation.revid_list) + (navigation.pagesize - 1)
    ) // navigation.pagesize

    def get_offset(offset):
        if (navigation.position + offset < 0) or (
            navigation.position + offset > navigation.count - 1
        ):
            return None
        return navigation.revid_list[navigation.position + offset]

    navigation.last_in_page_revid = get_offset(navigation.pagesize - 1)
    navigation.prev_page_revid = get_offset(-1 * navigation.pagesize)
    navigation.next_page_revid = get_offset(1 * navigation.pagesize)
    prev_page_revno = navigation.history.get_revno(navigation.prev_page_revid)
    next_page_revno = navigation.history.get_revno(navigation.next_page_revid)
    start_revno = navigation.history.get_revno(navigation.start_revid)

    params = {"filter_path": navigation.filter_path}
    if getattr(navigation, "query", None) is not None:
        params["q"] = navigation.query

    if getattr(navigation, "start_revid", None) is not None:
        params["start_revid"] = start_revno

    if navigation.prev_page_revid:
        navigation.prev_page_url = navigation.branch.context_url(
            [navigation.scan_url, prev_page_revno], **params
        )
    if navigation.next_page_revid:
        navigation.next_page_url = navigation.branch.context_url(
            [navigation.scan_url, next_page_revno], **params
        )


def directory_breadcrumbs(path, is_root, view):
    """
    Generate breadcrumb information from the directory path given

    The path given should be a path up to any branch that is currently being
    served

    Arguments:
    path -- The path to convert into breadcrumbs
    is_root -- Whether or not loggerhead is serving a branch at its root
    view -- The type of view we are showing (files, changes etc)
    """
    # Is our root directory itself a branch?
    if is_root:
        breadcrumbs = [
            {
                "dir_name": path,
                "path": "",
                "suffix": view,
            }
        ]
    else:
        # Create breadcrumb trail for the path leading up to the branch
        breadcrumbs = [
            {
                "dir_name": "(root)",
                "path": "",
                "suffix": "",
            }
        ]
        if path != "/":
            dir_parts = path.strip("/").split("/")
            for index, dir_name in enumerate(dir_parts):
                breadcrumbs.append(
                    {
                        "dir_name": dir_name,
                        "path": "/".join(dir_parts[: index + 1]),
                        "suffix": "",
                    }
                )
            # If we are not in the directory view, the last crumb is a branch,
            # so we need to specify a view
            if view != "directory":
                breadcrumbs[-1]["suffix"] = "/" + view
    return breadcrumbs


def branch_breadcrumbs(path, tree, view):
    """
    Generate breadcrumb information from the branch path given

    The path given should be a path that exists within a branch

    Arguments:
    path -- The path to convert into breadcrumbs
    tree -- Tree to get file information from
    view -- The type of view we are showing (files, changes etc)
    """
    dir_parts = path.strip("/").split("/")
    inner_breadcrumbs = []
    for index, dir_name in enumerate(dir_parts):
        inner_breadcrumbs.append(
            {
                "dir_name": dir_name,
                "path": "/".join(dir_parts[: index + 1]),
                "suffix": "/" + view,
            }
        )
    return inner_breadcrumbs


def decorator(unbound):
    def new_decorator(f):
        g = unbound(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g

    new_decorator.__name__ = unbound.__name__
    new_decorator.__doc__ = unbound.__doc__
    new_decorator.__dict__.update(unbound.__dict__)
    return new_decorator


@decorator
def lsprof(f):
    def _f(*a, **kw):
        import cPickle

        from .loggerhead.lsprof import profile

        z = time.time()
        ret, stats = profile(f, *a, **kw)
        log.debug(
            "Finished profiled %s in %d msec."
            % (f.__name__, int((time.time() - z) * 1000))
        )
        stats.sort()
        stats.freeze()
        now = time.time()
        msec = int(now * 1000) % 1000
        timestr = time.strftime("%Y%m%d%H%M%S", time.localtime(now)) + (
            "%03d" % (msec,)
        )
        filename = f.__name__ + "-" + timestr + ".lsprof"
        cPickle.dump(stats, open(filename, "w"), 2)
        return ret

    return _f


# just thinking out loud here...
#
# so, when browsing around, there are 5 pieces of context, most optional:
#     - current revid
#         current location along the navigation path (while browsing)
#     - starting revid (start_revid)
#         the current beginning of navigation (navigation continues back to
#         the original revision) -- this defines an 'alternate mainline'
#         when the user navigates into a branch.
#     - filter_path
#         if navigating the revisions that touched a file
#     - q (query)
#         if navigating the revisions that matched a search query
#     - remember
#         a previous revision to remember for future comparisons
#
# current revid is given on the url path.  the rest are optional components
# in the url params.
#
# other transient things can be set:
#     - compare_revid
#         to compare one revision to another, on /revision only
#     - sort
#         for re-ordering an existing page by different sort

t_context = threading.local()
_valid = ("start_revid", "filter_path", "q", "remember", "compare_revid", "sort")


def set_context(map):
    t_context.map = dict((k, v) for (k, v) in map.items() if k in _valid)


def get_context(**overrides):
    """
    Soon to be deprecated.


    return a context map that may be overridden by specific values passed in,
    but only contains keys from the list of valid context keys.

    if 'clear' is set, only the 'remember' context value will be added, and
    all other context will be omitted.
    """
    map = dict()
    if overrides.get("clear", False):
        map["remember"] = t_context.map.get("remember", None)
    else:
        map.update(t_context.map)
    overrides = dict((k, v) for (k, v) in overrides.items() if k in _valid)
    map.update(overrides)
    return map


class Reloader(object):
    """
    This class wraps all paste.reloader logic. All methods are @classmethod.
    """

    _reloader_environ_key = "PYTHON_RELOADER_SHOULD_RUN"

    @classmethod
    def _turn_sigterm_into_systemexit(cls):
        """
        Attempts to turn a SIGTERM exception into a SystemExit exception.
        """
        try:
            import signal
        except ImportError:
            return

        def handle_term(signo, frame):
            raise SystemExit

        signal.signal(signal.SIGTERM, handle_term)

    @classmethod
    def is_installed(cls):
        return os.environ.get(cls._reloader_environ_key)

    @classmethod
    def install(cls):
        from paste import reloader

        reloader.install(int(1))

    @classmethod
    def restart_with_reloader(cls):
        """Based on restart_with_monitor from paste.script.serve."""
        print("Starting subprocess with file monitor")
        while True:
            args = [sys.executable] + sys.argv
            new_environ = os.environ.copy()
            new_environ[cls._reloader_environ_key] = "true"
            proc = None
            try:
                try:
                    cls._turn_sigterm_into_systemexit()
                    proc = subprocess.Popen(args, env=new_environ)
                    exit_code = proc.wait()
                    proc = None
                except KeyboardInterrupt:
                    print("^C caught in monitor process")
                    return 1
            finally:
                if proc is not None and getattr(os, "kill", None) is not None:
                    import signal

                    try:
                        os.kill(proc.pid, signal.SIGTERM)
                    except (OSError, IOError):
                        pass

            # Reloader always exits with code 3; but if we are
            # a monitor, any exit code will restart
            if exit_code != 3:
                return exit_code
            print("-" * 20, "Restarting", "-" * 20)


def convert_file_errors(application):
    """WSGI wrapper to convert some file errors to Paste exceptions"""

    def new_application(environ, start_response):
        try:
            return application(environ, start_response)
        except (IOError, OSError) as e:
            import errno

            from paste import httpexceptions

            if e.errno == errno.ENOENT:
                raise httpexceptions.HTTPNotFound()
            elif e.errno == errno.EACCES:
                raise httpexceptions.HTTPForbidden()
            else:
                raise

    return new_application


def convert_to_json_ready(obj):
    if isinstance(obj, Container):
        d = obj.__dict__.copy()
        del d["_properties"]
        return d
    elif isinstance(obj, bytes):
        return obj.decode("UTF-8")
    elif isinstance(obj, datetime.datetime):
        return tuple(obj.utctimetuple())
    raise TypeError(repr(obj) + " is not JSON serializable")



# Copyright (C) 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#
"""WSGI apps tend to return body content as part of a HEAD request.

We should definitely not do that.
"""


class HeadMiddleware(object):
    """When we get a HEAD request, we should not return body content.

    WSGI defaults to just generating everything, and not paying attention to
    whether it is a GET or a HEAD request. It does that because of potential
    issues getting the Headers correct.

    This middleware works by just omitting the body if the request method is
    HEAD.
    """

    def __init__(self, app):
        self._wrapped_app = app
        self._real_environ = None
        self._real_start_response = None
        self._real_writer = None

    def noop_write(self, chunk):
        """We intentionally ignore all body content that is returned."""
        pass

    def start_response(self, status, response_headers, exc_info=None):
        if exc_info is None:
            self._real_writer = self._real_start_response(status, response_headers)
        else:
            self._real_writer = self._real_start_response(
                status, response_headers, exc_info
            )
        return self.noop_write

    def __call__(self, environ, start_response):
        self._real_environ = environ
        self._real_start_response = start_response
        if environ.get("REQUEST_METHOD", "GET") == "HEAD":
            result = self._wrapped_app(environ, self.start_response)
            for chunk in result:
                pass
        else:
            result = self._wrapped_app(environ, start_response)
            for chunk in result:
                yield chunk



# Copyright 2009, 2010, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA


# This file allows loggerhead to be treated as a plugin for bzr.
#
# XXX: Because loggerhead already contains a loggerhead directory, much of the
# code is going to appear loaded at breezy.plugins.loggerhead.loggerhead.
# This seems like the easiest thing, because breezy wants the top-level plugin
# directory to be the module, but when it's used as a library people expect
# the source directory to contain a directory called loggerhead.  -- mbp
# 20090123

"""Loggerhead web viewer for Bazaar branches.

This provides a new option "--http" to the "bzr serve" command, that
starts a web server to browse the contents of a branch.
"""

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

import sys

from packaging.version import Version

try:
    version_info = Version(importlib_metadata.version("loggerhead")).release
except importlib_metadata.PackageNotFoundError:
    # Support running tests from the build tree without installation.
    version_info = None

from breezy import commands
from breezy.transport import transport_server_registry

DEFAULT_HOST = "0.0.0.0"
DEFAULT_PORT = 8080
HELP = "Loggerhead, a web-based code viewer and server. (default port: %d)" % (
    DEFAULT_PORT,
)


def serve_http(transport, host=None, port=None, inet=None, client_timeout=None):
    # TODO: if we supported inet to pass requests in and respond to them,
    #       then it would be easier to test the full stack, but it probably
    #       means routing around paste.httpserver.serve which probably
    #       isn't testing the full stack
    from paste.httpexceptions import HTTPExceptionHandler
    from paste.httpserver import serve

    try:
        from .loggerhead.apps.transport import BranchesFromTransportRoot
        from .loggerhead.config import LoggerheadConfig
        from .loggerhead.__main__ import setup_logging
    except ImportError:
        from loggerhead.apps.transport import BranchesFromTransportRoot
        from loggerhead.config import LoggerheadConfig
        from loggerhead.__main__ import setup_logging

    if host is None:
        host = DEFAULT_HOST
    if port is None:
        port = DEFAULT_PORT
    argv = ["--host", host, "--port", str(port), "--", transport.base]
    if not transport.is_readonly():
        argv.insert(0, "--allow-writes")
    config = LoggerheadConfig(argv)
    setup_logging(config, init_logging=False, log_file=sys.stderr)
    app = BranchesFromTransportRoot(transport.base, config)
    app = HTTPExceptionHandler(app)
    serve(app, host=host, port=port)


transport_server_registry.register("http", serve_http, help=HELP)


class cmd_load_test_loggerhead(commands.Command):
    """Run a load test against a live loggerhead instance.

    Pass in the name of a script file to run. See loggerhead/load_test.py
    for a description of the file format.
    """

    hidden = True
    takes_args = ["filename"]

    def run(self, filename):
        try:
            from .loggerhead.loggerhead import load_test
        except ImportError:
            from loggerhead.loggerhead import load_test
        script = load_test.run_script(filename)
        for thread_id in sorted(script._threads):
            worker = script._threads[thread_id][0]
            for url, success, time in worker.stats:
                self.outf.write(" %5.3fs %s %s\n" % (time, str(success)[0], url))


commands.register_command(cmd_load_test_loggerhead)


def load_tests(loader, basic_tests, pattern):
    try:
        from .loggerhead.tests import test_suite
    except ImportError:
        from breezy.trace import mutter

        mutter("loggerhead tests not installed, not registering tests")
    else:
        basic_tests.addTest(test_suite())
    return basic_tests



#
# Copyright (C) 2008, 2009 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#
"""Cache the whole history data needed by loggerhead about a branch."""

import logging
import time

from breezy.revision import NULL_REVISION, is_null
from breezy.tsort import merge_sort


def _strip_NULL_ghosts(revision_graph):
    """
    Copied over from breezy meant as a temporary workaround for
    deprecated methods.
    """
    # Filter ghosts, and null:
    if NULL_REVISION in revision_graph:
        del revision_graph[NULL_REVISION]
    for key, parents in revision_graph.items():
        revision_graph[key] = tuple(
            parent for parent in parents if parent in revision_graph
        )
    return revision_graph


def compute_whole_history_data(branch):
    """Compute _rev_info and _rev_indices for a branch.

    See History.__doc__ for what these data structures mean.
    """
    z = time.time()

    last_revid = branch.last_revision()

    log = logging.getLogger("loggerhead.%s" % (branch.get_config().get_nickname(),))

    graph = branch.repository.get_graph()
    parent_map = dict(
        (key, value)
        for key, value in graph.iter_ancestry([last_revid])
        if value is not None
    )

    _revision_graph = _strip_NULL_ghosts(parent_map)

    _rev_info = []
    _rev_indices = {}

    if is_null(last_revid):
        _merge_sort = []
    else:
        _merge_sort = merge_sort(_revision_graph, last_revid, generate_revno=True)

    for info in _merge_sort:
        seq, revid, merge_depth, revno, end_of_merge = info
        revno_str = ".".join(str(n) for n in revno)
        parents = _revision_graph[revid]
        _rev_indices[revid] = len(_rev_info)
        _rev_info.append(
            [(seq, revid, merge_depth, revno_str, end_of_merge), (), parents]
        )

    for revid in _revision_graph.keys():
        if _rev_info[_rev_indices[revid]][0][2] == 0:
            continue
        for parent in _revision_graph[revid]:
            c = _rev_info[_rev_indices[parent]]
            if revid not in c[1]:
                c[1] = c[1] + (revid,)

    log.info("built revision graph cache: %.3f secs" % (time.time() - z,))

    return (_rev_info, _rev_indices)



#
# Copyright (C) 2008  Guillermo Gonzalez <guillo.gonzo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import traceback
from io import StringIO

from .. import util
from ..controllers import TemplatedBranchView


class ErrorUI(TemplatedBranchView):
    template_name = "error"

    def __init__(self, branch, exc_info):
        super(ErrorUI, self).__init__(branch, lambda: None)
        self.exc_info = exc_info

    def get_values(self, path, kwargs, response):
        exc_type, exc_object, exc_tb = self.exc_info
        description = StringIO()
        traceback.print_exception(exc_type, exc_object, None, file=description)
        directory_breadcrumbs = util.directory_breadcrumbs(
            self._branch.friendly_name, self._branch.is_root, "changes"
        )
        return {
            "branch": self._branch,
            "error_title": (
                "An unexpected error occurred whileprocessing the request:"
            ),
            "error_description": description.getvalue(),
            "directory_breadcrumbs": directory_breadcrumbs,
        }



# Copyright 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#

"""Tests for the load testing code."""

import socket
import threading
import time
from queue import Empty

from breezy import tests
from breezy.tests import http_server

from .. import load_test

empty_script = """{
    "parameters": {},
    "requests": []
}"""


class TestRequestDescription(tests.TestCase):
    def test_init_from_dict(self):
        rd = load_test.RequestDescription({"thread": "10", "relpath": "/foo"})
        self.assertEqual("10", rd.thread)
        self.assertEqual("/foo", rd.relpath)

    def test_default_thread_is_1(self):
        rd = load_test.RequestDescription({"relpath": "/bar"})
        self.assertEqual("1", rd.thread)
        self.assertEqual("/bar", rd.relpath)


_cur_time = time.time()


def one_sec_timer():
    """Every time this timer is called, it increments by 1 second."""
    global _cur_time
    _cur_time += 1.0
    return _cur_time


class NoopRequestWorker(load_test.RequestWorker):
    # Every call to _timer will increment by one
    _timer = staticmethod(one_sec_timer)

    # Ensure that process never does anything
    def process(self, url):
        return True


class TestRequestWorkerInfrastructure(tests.TestCase):
    """Tests various infrastructure bits, without doing actual requests."""

    def test_step_next_tracks_time(self):
        rt = NoopRequestWorker("id")
        rt.queue.put("item")
        rt.step_next()
        self.assertTrue(rt.queue.empty())
        self.assertEqual([("item", True, 1.0)], rt.stats)

    def test_step_multiple_items(self):
        rt = NoopRequestWorker("id")
        rt.queue.put("item")
        rt.step_next()
        rt.queue.put("next-item")
        rt.step_next()
        self.assertTrue(rt.queue.empty())
        self.assertEqual([("item", True, 1.0), ("next-item", True, 1.0)], rt.stats)

    def test_step_next_does_nothing_for_noop(self):
        rt = NoopRequestWorker("id")
        rt.queue.put("item")
        rt.step_next()
        rt.queue.put("<noop>")
        rt.step_next()
        self.assertEqual([("item", True, 1.0)], rt.stats)

    def test_step_next_will_timeout(self):
        # We don't want step_next to block forever
        rt = NoopRequestWorker("id", blocking_time=0.001)
        self.assertRaises(Empty, rt.step_next)

    def test_run_stops_for_stop_event(self):
        rt = NoopRequestWorker("id", blocking_time=0.001, _queue_size=2)
        rt.queue.put("item1")
        rt.queue.put("item2")
        event = threading.Event()
        t = threading.Thread(target=rt.run, args=(event,))
        t.start()
        # Wait for the queue to be processed
        rt.queue.join()
        # Now we can queue up another one, and wait for it
        rt.queue.put("item3")
        rt.queue.join()
        # Now set the stopping event
        event.set()
        # Add another item to the queue, which might get processed, but the
        # next item won't
        rt.queue.put("item4")
        rt.queue.put("item5")
        t.join()
        self.assertEqual(
            [("item1", True, 1.0), ("item2", True, 1.0), ("item3", True, 1.0)],
            rt.stats[:3],
        )
        # The last event might be item4 or might be item3, the important thing
        # is that even though there are still queued events, we won't
        # process anything past the first
        self.assertNotEqual("item5", rt.stats[-1][0])


class TestRequestWorker(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.transport_readonly_server = http_server.HttpServer

    def test_request_items(self):
        rt = load_test.RequestWorker("id", blocking_time=0.01, _queue_size=2)
        self.build_tree(["file1", "file2"])
        readonly_url1 = self.get_readonly_url("file1")
        self.assertStartsWith(readonly_url1, "http://")
        readonly_url2 = self.get_readonly_url("file2")
        rt.queue.put(readonly_url1)
        rt.queue.put(readonly_url2)
        rt.step_next()
        rt.step_next()
        self.assertEqual(readonly_url1, rt.stats[0][0])
        self.assertEqual(readonly_url2, rt.stats[1][0])

    def test_request_nonexistant_items(self):
        rt = load_test.RequestWorker("id", blocking_time=0.01, _queue_size=2)
        readonly_url1 = self.get_readonly_url("no-file1")
        rt.queue.put(readonly_url1)
        rt.step_next()
        self.assertEqual(readonly_url1, rt.stats[0][0])
        self.assertEqual(False, rt.stats[0][1])

    def test_no_server(self):
        s = socket.socket()
        # Bind to a port, but don't listen on it
        s.bind(("localhost", 0))
        url = "http://%s:%s/path" % s.getsockname()
        rt = load_test.RequestWorker("id", blocking_time=0.01, _queue_size=2)
        rt.queue.put(url)
        rt.step_next()
        self.assertEqual((url, False), rt.stats[0][:2])


class NoActionScript(load_test.ActionScript):
    _thread_class = NoopRequestWorker
    _default_blocking_timeout = 0.01


class TestActionScriptInfrastructure(tests.TestCase):
    def test_parse_requires_parameters_and_requests(self):
        self.assertRaises(ValueError, load_test.ActionScript.parse, "")
        self.assertRaises(ValueError, load_test.ActionScript.parse, "{}")
        self.assertRaises(
            ValueError, load_test.ActionScript.parse, '{"parameters": {}}'
        )
        self.assertRaises(ValueError, load_test.ActionScript.parse, '{"requests": []}')
        load_test.ActionScript.parse(
            '{"parameters": {}, "requests": [], "comment": "section"}'
        )
        script = load_test.ActionScript.parse(empty_script)
        self.assertIsNot(None, script)

    def test_parse_default_base_url(self):
        script = load_test.ActionScript.parse(empty_script)
        self.assertEqual("http://localhost:8080", script.base_url)

    def test_parse_find_base_url(self):
        script = load_test.ActionScript.parse(
            '{"parameters": {"base_url": "http://example.com"}, "requests": []}'
        )
        self.assertEqual("http://example.com", script.base_url)

    def test_parse_default_blocking_timeout(self):
        script = load_test.ActionScript.parse(empty_script)
        self.assertEqual(60.0, script.blocking_timeout)

    def test_parse_find_blocking_timeout(self):
        script = load_test.ActionScript.parse(
            '{"parameters": {"blocking_timeout": 10.0}, "requests": []}'
        )
        self.assertEqual(10.0, script.blocking_timeout)

    def test_parse_finds_requests(self):
        script = load_test.ActionScript.parse(
            '{"parameters": {}, "requests": ['
            ' {"relpath": "/foo"},'
            ' {"relpath": "/bar"}'
            " ]}"
        )
        self.assertEqual(2, len(script._requests))
        self.assertEqual("/foo", script._requests[0].relpath)
        self.assertEqual("/bar", script._requests[1].relpath)

    def test__get_worker(self):
        script = NoActionScript()
        # If an id is found, then we should create it
        self.assertEqual({}, script._threads)
        worker = script._get_worker("id")
        self.assertIn("id", script._threads)
        # We should have set the blocking timeout
        self.assertEqual(script.blocking_timeout / 10.0, worker.blocking_time)

        # Another request will return the identical object
        self.assertIs(worker, script._get_worker("id"))

        # And the stop event will stop the thread
        script.stop_and_join()

    def test__full_url(self):
        script = NoActionScript()
        self.assertEqual("http://localhost:8080/path", script._full_url("/path"))
        self.assertEqual(
            "http://localhost:8080/path/to/foo", script._full_url("/path/to/foo")
        )
        script.base_url = "http://example.com"
        self.assertEqual(
            "http://example.com/path/to/foo", script._full_url("/path/to/foo")
        )
        script.base_url = "http://example.com/base"
        self.assertEqual(
            "http://example.com/base/path/to/foo", script._full_url("/path/to/foo")
        )
        script.base_url = "http://example.com"
        self.assertEqual("http://example.com:8080/path", script._full_url(":8080/path"))

    def test_single_threaded(self):
        script = NoActionScript.parse("""{
            "parameters": {"base_url": ""},
            "requests": [
                {"thread": "1", "relpath": "first"},
                {"thread": "1", "relpath": "second"},
                {"thread": "1", "relpath": "third"},
                {"thread": "1", "relpath": "fourth"}
            ]}""")
        script.run()
        worker = script._get_worker("1")
        self.assertEqual(
            ["first", "second", "third", "fourth"], [s[0] for s in worker.stats]
        )

    def test_two_threads(self):
        script = NoActionScript.parse("""{
            "parameters": {"base_url": ""},
            "requests": [
                {"thread": "1", "relpath": "first"},
                {"thread": "2", "relpath": "second"},
                {"thread": "1", "relpath": "third"},
                {"thread": "2", "relpath": "fourth"}
            ]}""")
        script.run()
        worker = script._get_worker("1")
        self.assertEqual(["first", "third"], [s[0] for s in worker.stats])
        worker = script._get_worker("2")
        self.assertEqual(["second", "fourth"], [s[0] for s in worker.stats])


class TestActionScriptIntegration(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.transport_readonly_server = http_server.HttpServer

    def test_full_integration(self):
        self.build_tree(["first", "second", "third", "fourth"])
        url = self.get_readonly_url()
        script = load_test.ActionScript.parse(
            """{{
            "parameters": {{"base_url": "{}", "blocking_timeout": 2.0}},
            "requests": [
                {{"thread": "1", "relpath": "first"}},
                {{"thread": "2", "relpath": "second"}},
                {{"thread": "1", "relpath": "no-this"}},
                {{"thread": "2", "relpath": "or-this"}},
                {{"thread": "1", "relpath": "third"}},
                {{"thread": "2", "relpath": "fourth"}}
            ]}}""".format(url)
        )
        script.run()
        worker = script._get_worker("1")
        self.assertEqual(
            [("first", True), ("no-this", False), ("third", True)],
            [(s[0].rsplit("/", 1)[1], s[1]) for s in worker.stats],
        )
        worker = script._get_worker("2")
        self.assertEqual(
            [("second", True), ("or-this", False), ("fourth", True)],
            [(s[0].rsplit("/", 1)[1], s[1]) for s in worker.stats],
        )


class TestRunScript(tests.TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.transport_readonly_server = http_server.HttpServer

    def test_run_script(self):
        self.build_tree(["file1", "file2", "file3", "file4"])
        url = self.get_readonly_url()
        self.build_tree_contents(
            [
                (
                    "localhost.script",
                    """{{
    "parameters": {{"base_url": "{}", "blocking_timeout": 0.1}},
    "requests": [
        {{"thread": "1", "relpath": "file1"}},
        {{"thread": "2", "relpath": "file2"}},
        {{"thread": "1", "relpath": "file3"}},
        {{"thread": "2", "relpath": "file4"}}
    ]
}}""".format(url),
                )
            ]
        )
        script = load_test.run_script("localhost.script")
        worker = script._threads["1"][0]
        self.assertEqual(
            [("file1", True), ("file3", True)],
            [(s[0].rsplit("/", 1)[1], s[1]) for s in worker.stats],
        )
        worker = script._threads["2"][0]
        self.assertEqual(
            [("file2", True), ("file4", True)],
            [(s[0].rsplit("/", 1)[1], s[1]) for s in worker.stats],
        )



#
# Copyright (C) 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
"""Configuration tools for Loggerhead."""

import sys
import tempfile
from optparse import OptionParser

from breezy import config

_temporary_sql_dir = None


def _get_temporary_sql_dir():
    global _temporary_sql_dir
    if _temporary_sql_dir is None:
        _temporary_sql_dir = tempfile.mkdtemp(prefix="loggerhead-cache-")
    return _temporary_sql_dir


def command_line_parser():
    parser = OptionParser("%prog [options] <path>")
    parser.set_defaults(
        user_dirs=False,
        show_version=False,
        log_folder=None,
        use_cdn=False,
        sql_dir=None,
        allow_writes=False,
        export_tarballs=True,
    )
    parser.add_option(
        "--user-dirs", action="store_true", help="Serve user directories as ~user."
    )
    parser.add_option(
        "--trunk-dir",
        metavar="DIR",
        help="The directory that contains the trunk branches.",
    )
    parser.add_option(
        "--port",
        dest="user_port",
        help=("Port Loggerhead should listen on (defaults to 8080)."),
    )
    parser.add_option(
        "--host", dest="user_host", help="Host Loggerhead should listen on."
    )
    parser.add_option(
        "--protocol",
        dest="protocol",
        help=("Protocol to use: http, scgi, fcgi, ajp(defaults to http)."),
    )
    parser.add_option(
        "--log-level",
        default=None,
        action="callback",
        callback=_optparse_level_to_int_level,
        type="string",
        help="Set the verbosity of logging. Can either"
        " be set to a numeric or string"
        " (eg, 10=debug, 30=warning)",
    )
    parser.add_option(
        "--memory-profile",
        action="store_true",
        help="Profile the memory usage using Dozer.",
    )
    parser.add_option("--prefix", dest="user_prefix", help="Specify host prefix.")
    parser.add_option(
        "--profile",
        action="store_true",
        help="Generate callgrind profile data to %d-stats.callgrind on each request.",
    )
    parser.add_option(
        "--reload",
        action="store_true",
        help="Restarts the application when changing python"
        " files. Only used for development purposes.",
    )
    parser.add_option("--log-folder", help="The directory to place log files in.")
    parser.add_option(
        "--version",
        action="store_true",
        dest="show_version",
        help="Print the software version and exit",
    )
    parser.add_option(
        "--use-cdn",
        action="store_true",
        dest="use_cdn",
        help="Serve jQuery from Google's CDN",
    )
    parser.add_option(
        "--cache-dir", dest="sql_dir", help="The directory to place the SQL cache in"
    )
    parser.add_option(
        "--allow-writes",
        action="store_true",
        help="Allow writing to the Bazaar server.",
    )
    parser.add_option(
        "--export-tarballs",
        action="store_true",
        help="Allow exporting revisions to tarballs.",
    )
    return parser


_log_levels = {
    "debug": 10,
    "info": 20,
    "warning": 30,
    "error": 40,
    "critical": 50,
}


def _optparse_level_to_int_level(option, opt_str, value, parser):
    parser.values.log_level = _level_to_int_level(value)


def _level_to_int_level(value):
    """Convert a string level to an integer value."""
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        pass
    return _log_levels[value.lower()]


class LoggerheadConfig(object):
    """A configuration object."""

    def __init__(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]
        self._parser = command_line_parser()
        self._options, self._args = self._parser.parse_args(argv)

        sql_dir = self.get_option("sql_dir")
        if sql_dir is None:
            sql_dir = _get_temporary_sql_dir()
        self.SQL_DIR = sql_dir

    def get_option(self, option):
        """Get the value for the config option, either
        from ~/.config/breezy/breezy.conf or from the command line.
        All loggerhead-specific settings start with 'http_'
        """
        global_config = config.GlobalConfig().get_user_option("http_" + option)
        cmd_config = getattr(self._options, option)
        if global_config is not None and (cmd_config is None or cmd_config is False):
            return global_config
        else:
            return cmd_config

    def get_log_level(self):
        opt = self.get_option("log_level")
        return _level_to_int_level(opt)

    def get_arg(self, index):
        """Get an arg from the arg list."""
        return self._args[index]

    def print_help(self):
        """Wrapper around OptionParser.print_help."""
        return self._parser.print_help()

    @property
    def arg_count(self):
        """Return the number of args from the option parser."""
        return len(self._args)



# Copyright (C) 2006-2011 Canonical Ltd.
#                     (Authored by Martin Albisetti <argentina@gmail.com>)
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
# Copyright (C) 2005  Jake Edge <jake@edge2.net>
# Copyright (C) 2005  Matt Mackall <mpm@selenic.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

#
# This file (and many of the web templates) contains work based on the
# "bazaar-webserve" project by Goffredo Baroncelli, which is in turn based
# on "hgweb" by Jake Edge and Matt Mackall.
#


import bisect
import datetime
import logging
import re
import textwrap
import threading

import breezy.branch
import breezy.delta
import breezy.errors
import breezy.foreign
import breezy.osutils
import breezy.revision
from breezy import tag

try:
    from breezy.transport import NoSuchFile
except ImportError:
    from breezy.errors import NoSuchFile

from . import search, util
from .wholehistory import compute_whole_history_data


def is_branch(folder):
    try:
        breezy.branch.Branch.open(folder)
        return True
    except breezy.errors.NotBranchError:
        return False


def clean_message(message):
    """Clean up a commit message and return it and a short (1-line) version.

    Commit messages that are long single lines are reflowed using the textwrap
    module (Robey, the original author of this code, apparently favored this
    style of message).
    """
    message = message.lstrip().splitlines()

    if len(message) == 1:
        message = textwrap.wrap(message[0])

    if len(message) == 0:
        # We can end up where when (a) the commit message was empty or (b)
        # when the message consisted entirely of whitespace, in which case
        # textwrap.wrap() returns an empty list.
        return [""], ""

    # Make short form of commit message.
    short_message = message[0]
    if len(short_message) > 60:
        short_message = short_message[:60] + "..."

    return message, short_message


def rich_filename(path, kind):
    if kind == "directory":
        path += "/"
    if kind == "symlink":
        path += "@"
    return path


class _RevListToTimestamps(object):
    """This takes a list of revisions, and allows you to bisect by date"""

    __slots__ = ["revid_list", "repository"]

    def __init__(self, revid_list, repository):
        self.revid_list = revid_list
        self.repository = repository

    def __getitem__(self, index):
        """Get the date of the index'd item"""
        return datetime.datetime.fromtimestamp(
            self.repository.get_revision(self.revid_list[index]).timestamp
        )

    def __len__(self):
        return len(self.revid_list)


class FileChangeReporter(object):
    def __init__(self, old_tree, new_tree):
        self.added = []
        self.modified = []
        self.renamed = []
        self.removed = []
        self.text_changes = []
        self.old_tree = old_tree
        self.new_tree = new_tree

    def revid(self, tree, path):
        if path is None:
            return breezy.revision.NULL_REVISION
        try:
            return tree.get_file_revision(path)
        except NoSuchFile:
            return breezy.revision.NULL_REVISION

    def report(self, paths, versioned, renamed, copied, modified, exe_change, kind):
        return self._report(
            paths, versioned, renamed, copied, modified, exe_change, kind
        )

    def _report(self, paths, versioned, renamed, copied, modified, exe_change, kind):
        if modified not in ("unchanged", "kind changed"):
            if versioned == "removed":
                filename = rich_filename(paths[0], kind[0])
            else:
                filename = rich_filename(paths[1], kind[1])
            self.text_changes.append(
                util.Container(
                    filename=filename,
                    old_revision=self.revid(self.old_tree, paths[0]),
                    new_revision=self.revid(self.new_tree, paths[1]),
                )
            )
        if versioned == "added":
            self.added.append(
                util.Container(filename=rich_filename(paths[1], kind), kind=kind[1])
            )
        elif versioned == "removed":
            self.removed.append(
                util.Container(filename=rich_filename(paths[0], kind), kind=kind[0])
            )
        elif renamed:
            self.renamed.append(
                util.Container(
                    old_filename=rich_filename(paths[0], kind[0]),
                    new_filename=rich_filename(paths[1], kind[1]),
                    text_modified=modified == "modified",
                    exe_change=exe_change,
                )
            )
        else:
            self.modified.append(
                util.Container(
                    filename=rich_filename(paths[1], kind),
                    text_modified=modified == "modified",
                    exe_change=exe_change,
                )
            )


# The lru_cache is not thread-safe, so we need a lock around it for
# all threads.
rev_info_memory_cache_lock = threading.RLock()


class RevInfoMemoryCache(object):
    """A store that validates values against the revids they were stored with.

    We use a unique key for each branch.

    The reason for not just using the revid as the key is so that when a new
    value is provided for a branch, we replace the old value used for the
    branch.

    There is another implementation of the same interface in
    loggerhead.changecache.RevInfoDiskCache.
    """

    def __init__(self, cache):
        self._cache = cache

    def get(self, key, revid):
        """Return the data associated with `key`, subject to a revid check.

        If a value was stored under `key`, with the same revid, return it.
        Otherwise return None.
        """
        rev_info_memory_cache_lock.acquire()
        try:
            cached = self._cache.get(key)
        finally:
            rev_info_memory_cache_lock.release()
        if cached is None:
            return None
        stored_revid, data = cached
        if revid == stored_revid:
            return data
        else:
            return None

    def set(self, key, revid, data):
        """Store `data` under `key`, to be checked against `revid` on get()."""
        rev_info_memory_cache_lock.acquire()
        try:
            self._cache[key] = (revid, data)
        finally:
            rev_info_memory_cache_lock.release()


# Used to store locks that prevent multiple threads from building a
# revision graph for the same branch at the same time, because that can
# cause severe performance issues that are so bad that the system seems
# to hang.
revision_graph_locks = {}
revision_graph_check_lock = threading.Lock()


class History(object):
    """Decorate a branch to provide information for rendering.

    History objects are expected to be short lived -- when serving a request
    for a particular branch, open it, read-lock it, wrap a History object
    around it, serve the request, throw the History object away, unlock the
    branch and throw it away.

    :ivar _rev_info: A list of information about revisions.  This is by far
        the most cryptic data structure in loggerhead.  At the top level, it
        is a list of 3-tuples [(merge-info, where-merged, parents)].
        `merge-info` is (seq, revid, merge_depth, revno_str, end_of_merge) --
        like a merged sorted list, but the revno is stringified.
        `where-merged` is a tuple of revisions that have this revision as a
        non-lefthand parent.  Finally, `parents` is just the usual list of
        parents of this revision.
    :ivar _rev_indices: A dictionary mapping each revision id to the index of
        the information about it in _rev_info.
    :ivar _revno_revid: A dictionary mapping stringified revnos to revision
        ids.
    """

    def _load_whole_history_data(self, caches, cache_key):
        """Set the attributes relating to the whole history of the branch.

        :param caches: a list of caches with interfaces like
            `RevInfoMemoryCache` and be ordered from fastest to slowest.
        :param cache_key: the key to use with the caches.
        """
        self._rev_indices = None
        self._rev_info = None

        missed_caches = []

        def update_missed_caches():
            for cache in missed_caches:
                cache.set(cache_key, self.last_revid, self._rev_info)

        # Theoretically, it's possible for two threads to race in creating
        # the Lock() object for their branch, so we put a lock around
        # creating the per-branch Lock().
        revision_graph_check_lock.acquire()
        try:
            if cache_key not in revision_graph_locks:
                revision_graph_locks[cache_key] = threading.Lock()
        finally:
            revision_graph_check_lock.release()

        revision_graph_locks[cache_key].acquire()
        try:
            for cache in caches:
                data = cache.get(cache_key, self.last_revid)
                if data is not None:
                    self._rev_info = data
                    update_missed_caches()
                    break
                else:
                    missed_caches.append(cache)
            else:
                whole_history_data = compute_whole_history_data(self._branch)
                self._rev_info, self._rev_indices = whole_history_data
                update_missed_caches()
        finally:
            revision_graph_locks[cache_key].release()

        if self._rev_indices is not None:
            self._revno_revid = {}
            for (_, revid, _, revno_str, _), _, _ in self._rev_info:
                self._revno_revid[revno_str] = revid
        else:
            self._revno_revid = {}
            self._rev_indices = {}
            for (seq, revid, _, revno_str, _), _, _ in self._rev_info:
                self._rev_indices[revid] = seq
                self._revno_revid[revno_str] = revid

    def __init__(
        self, branch, whole_history_data_cache, revinfo_disk_cache=None, cache_key=None
    ):
        assert branch.is_locked(), (
            "Can only construct a History object with a read-locked branch."
        )
        self._branch = branch
        self._branch_tags = None
        self._inventory_cache = {}
        self._branch_nick = self._branch.get_config().get_nickname()
        self.log = logging.getLogger("loggerhead.%s" % (self._branch_nick,))

        self.last_revid = branch.last_revision()

        caches = [RevInfoMemoryCache(whole_history_data_cache)]
        if revinfo_disk_cache:
            caches.append(revinfo_disk_cache)
        self._load_whole_history_data(caches, cache_key)

    @property
    def has_revisions(self):
        return not breezy.revision.is_null(self.last_revid)

    def get_config(self):
        return self._branch.get_config()

    def get_revno(self, revid):
        if revid not in self._rev_indices:
            # ghost parent?
            return "unknown"
        seq = self._rev_indices[revid]
        revno = self._rev_info[seq][0][3]
        return revno

    def get_revids_from(self, revid_list, start_revid):
        """
        Yield the mainline (wrt start_revid) revisions that merged each
        revid in revid_list.
        """
        if revid_list is None:
            # Just yield the mainline, starting at start_revid
            revid = start_revid
            is_null = breezy.revision.is_null
            while not is_null(revid):
                yield revid
                parents = self._rev_info[self._rev_indices[revid]][2]
                if not parents:
                    return
                revid = parents[0]
            return
        revid_set = set(revid_list)
        revid = start_revid

        def introduced_revisions(revid):
            r = set([revid])
            seq = self._rev_indices[revid]
            md = self._rev_info[seq][0][2]
            i = seq + 1
            while i < len(self._rev_info) and self._rev_info[i][0][2] > md:
                r.add(self._rev_info[i][0][1])
                i += 1
            return r

        while revid_set:
            if breezy.revision.is_null(revid):
                return
            rev_introduced = introduced_revisions(revid)
            matching = rev_introduced.intersection(revid_set)
            if matching:
                # We don't need to look for these anymore.
                revid_set.difference_update(matching)
                yield revid
            parents = self._rev_info[self._rev_indices[revid]][2]
            if len(parents) == 0:
                return
            revid = parents[0]

    def get_short_revision_history_by_fileid(self, file_id):
        # FIXME: would be awesome if we could get, for a folder, the list of
        # revisions where items within that folder changed.i
        # TODO(jelmer): Avoid versionedfile-specific texts
        possible_keys = [(file_id, revid) for revid in self._rev_indices]
        get_parent_map = self._branch.repository.texts.get_parent_map
        # We chunk the requests as this works better with GraphIndex.
        # See _filter_revisions_touching_file_id in breezy/log.py
        # for more information.
        revids = []
        chunk_size = 1000
        for start in range(0, len(possible_keys), chunk_size):
            next_keys = possible_keys[start : start + chunk_size]
            revids += [k[1] for k in get_parent_map(next_keys)]
        del possible_keys, next_keys
        return revids

    def get_revision_history_since(self, revid_list, date):
        # if a user asks for revisions starting at 01-sep, they mean inclusive,
        # so start at midnight on 02-sep.
        date = date + datetime.timedelta(days=1)
        # our revid list is sorted in REVERSE date order,
        # so go thru some hoops here...
        revid_list.reverse()
        index = bisect.bisect(
            _RevListToTimestamps(revid_list, self._branch.repository), date
        )
        if index == 0:
            return []
        revid_list.reverse()
        index = -index
        return revid_list[index:]

    def get_search_revid_list(self, query, revid_list):
        """
        given a "quick-search" query, try a few obvious possible meanings:

            - revision id or # ("128.1.3")
            - date (US style "mm/dd/yy", earth style "dd-mm-yy", or \
iso style "yyyy-mm-dd")
            - comment text as a fallback

        and return a revid list that matches.
        """
        # FIXME: there is some silliness in this action.  we have to look up
        # all the relevant changes (time-consuming) only to return a list of
        # revids which will be used to fetch a set of changes again.

        # if they entered a revid, just jump straight there;
        # ignore the passed-in revid_list
        revid = self.fix_revid(query)
        if revid is not None:
            changes = self.get_changes([revid])
            if (changes is not None) and (len(changes) > 0):
                return [revid]

        date = None
        m = self.us_date_re.match(query)
        if m is not None:
            date = datetime.datetime(
                util.fix_year(int(m.group(3))), int(m.group(1)), int(m.group(2))
            )
        else:
            m = self.earth_date_re.match(query)
            if m is not None:
                date = datetime.datetime(
                    util.fix_year(int(m.group(3))), int(m.group(2)), int(m.group(1))
                )
            else:
                m = self.iso_date_re.match(query)
                if m is not None:
                    date = datetime.datetime(
                        util.fix_year(int(m.group(1))), int(m.group(2)), int(m.group(3))
                    )
        if date is not None:
            if revid_list is None:
                # if no limit to the query was given,
                # search only the direct-parent path.
                revid_list = list(self.get_revids_from(None, self.last_revid))
            return self.get_revision_history_since(revid_list, date)

    revno_re = re.compile(r"^[\d\.]+$")
    # the date regex are without a final '$' so that queries like
    # "2006-11-30 12:15" still mostly work.  (i think it's better to give
    # them 90% of what they want instead of nothing at all.)
    us_date_re = re.compile(r"^(\d{1,2})/(\d{1,2})/(\d\d(\d\d?))")
    earth_date_re = re.compile(r"^(\d{1,2})-(\d{1,2})-(\d\d(\d\d?))")
    iso_date_re = re.compile(r"^(\d\d\d\d)-(\d\d)-(\d\d)")

    def fix_revid(self, revid):
        # if a "revid" is actually a dotted revno, convert it to a revid
        if revid is None:
            return revid
        if not isinstance(revid, str):
            raise TypeError(revid)
        if revid == "head:":
            return self.last_revid
        try:
            if self.revno_re.match(revid):
                revid = self._revno_revid[revid]
        except KeyError:
            raise breezy.errors.NoSuchRevision(self._branch_nick, revid)
        if not isinstance(revid, bytes):
            revid = revid.encode("utf-8")
        return revid

    @staticmethod
    def _iterate_sufficiently(iterable, stop_at, extra_rev_count):
        """Return a list of iterable.

        If extra_rev_count is None, fully consume iterable.
        Otherwise, stop at 'stop_at' + extra_rev_count.

        Example:
          iterate until you find stop_at, then iterate 10 more times.
        """
        if extra_rev_count is None:
            return list(iterable)
        result = []
        found = False
        for n in iterable:
            result.append(n)
            if n == stop_at:
                found = True
                break
        if found:
            for count, n in enumerate(iterable):
                if count >= extra_rev_count:
                    break
                result.append(n)
        return result

    def _get_file_view(self, revid, file_id):
        """
        Given a revid and optional path, return a (revlist, revid) for
        navigation through the current scope: from the revid (or the latest
        revision) back to the original revision.

        If file_id is None, the entire revision history is the list scope.
        """
        if revid is None:
            revid = self.last_revid
        if file_id is not None:
            revlist = list(self.get_short_revision_history_by_fileid(file_id))
            revlist = self.get_revids_from(revlist, revid)
        else:
            revlist = self.get_revids_from(None, revid)
        return revlist

    def get_view(self, revid, start_revid, path, query=None, extra_rev_count=None):
        """
        use the URL parameters (revid, start_revid, path, and query) to
        determine the revision list we're viewing (start_revid, path, query)
        and where we are in it (revid).

            - if a query is given, we're viewing query results.
            - if a path is given, we're viewing revisions for a specific
              file.
            - if a start_revid is given, we're viewing the branch from a
              specific revision up the tree.
            - if extra_rev_count is given, find the view from start_revid =>
              revid, and continue an additional 'extra_rev_count'. If not
              given, then revid_list will contain the full history of
              start_revid

        these may be combined to view revisions for a specific file, from
        a specific revision, with a specific search query.

        returns a new (revid, start_revid, revid_list) where:

            - revid: current position within the view
            - start_revid: starting revision of this view
            - revid_list: list of revision ids for this view

        path and query are never changed so aren't returned, but they may
        contain vital context for future url navigation.
        """
        if start_revid is None:
            start_revid = self.last_revid

        if query is None:
            repo = self._branch.repository
            if path is not None:
                file_id = repo.revision_tree(start_revid).path2id(path)
            else:
                file_id = None
            revid_list = self._get_file_view(start_revid, file_id)
            revid_list = self._iterate_sufficiently(revid_list, revid, extra_rev_count)
            if revid is None:
                revid = start_revid
            if revid not in revid_list:
                # if the given revid is not in the revlist, use a revlist that
                # starts at the given revid.
                revid_list = self._get_file_view(revid, file_id)
                revid_list = self._iterate_sufficiently(
                    revid_list, revid, extra_rev_count
                )
                start_revid = revid
            return revid, start_revid, revid_list
        else:
            file_id = None

        # potentially limit the search
        if file_id is not None:
            revid_list = self._get_file_view(start_revid, file_id)
        else:
            revid_list = None
        revid_list = search.search_revisions(self._branch, query)
        if revid_list and len(revid_list) > 0:
            if revid not in revid_list:
                revid = revid_list[0]
            return revid, start_revid, revid_list
        else:
            # XXX: This should return a message saying that the search could
            # not be completed due to either missing the plugin or missing a
            # search index.
            return None, None, []

    def revision_tree(self, revid):
        return self._branch.repository.revision_tree(revid)

    def file_exists(self, revid, path):
        if (len(path) > 0) and not path.startswith("/"):
            path = "/" + path
        try:
            return self.revision_tree(revid).has_filename(path)
        except breezy.errors.NoSuchRevision:
            return False

    def get_merge_point_list(self, revid):
        """
        Return the list of revids that have merged this node.
        """
        if "." not in self.get_revno(revid):
            return []

        merge_point = []
        nexts = [revid]
        while nexts:
            revid = nexts.pop()
            children = self._rev_info[self._rev_indices[revid]][1]
            for child in children:
                child_parents = self._rev_info[self._rev_indices[child]][2]
                if child_parents[0] == revid:
                    nexts.append(child)
                else:
                    merge_point.append(child)
        return merge_point

    def simplify_merge_point_list(self, revids):
        """if a revision is already merged, don't show further merge points"""
        d = {}
        for revid in revids:
            revno = self.get_revno(revid)
            revnol = revno.split(".")
            revnos = ".".join(revnol[:-2])
            revnolast = int(revnol[-1])
            if revnos in d:
                m = d[revnos][0]
                if revnolast < m:
                    d[revnos] = (revnolast, revid)
            else:
                d[revnos] = (revnolast, revid)

        return [revid for (_, revid) in d.values()]

    def add_branch_nicks(self, change):
        """
        given a 'change', fill in the branch nicks on all parents and merge
        points.
        """
        fetch_set = set()
        for p in change.parents:
            fetch_set.add(p.revid)
        for p in change.merge_points:
            fetch_set.add(p.revid)
        p_changes = self.get_changes(list(fetch_set))
        p_change_dict = {c.revid: c for c in p_changes}
        for p in change.parents:
            if p.revid in p_change_dict:
                p.branch_nick = p_change_dict[p.revid].branch_nick
            else:
                p.branch_nick = "(missing)"
        for p in change.merge_points:
            if p.revid in p_change_dict:
                p.branch_nick = p_change_dict[p.revid].branch_nick
            else:
                p.branch_nick = "(missing)"

    def get_changes(self, revid_list):
        """Return a list of changes objects for the given revids.

        Revisions not present and NULL_REVISION will be ignored.
        """
        for revid in revid_list:
            if not isinstance(revid, bytes):
                raise TypeError(revid_list)
        changes = self.get_changes_uncached(revid_list)
        if len(changes) == 0:
            return changes

        # some data needs to be recalculated each time, because it may
        # change as new revisions are added.
        for change in changes:
            merge_revids = self.simplify_merge_point_list(
                self.get_merge_point_list(change.revid)
            )
            change.merge_points = [
                util.Container(revid=r, revno=self.get_revno(r)) for r in merge_revids
            ]
            if len(change.parents) > 0:
                change.parents = [
                    util.Container(revid=r, revno=self.get_revno(r))
                    for r in change.parents
                ]
            change.revno = self.get_revno(change.revid)

        parity = 0
        for change in changes:
            change.parity = parity
            parity ^= 1

        return changes

    def get_changes_uncached(self, revid_list):
        # FIXME: deprecated method in getting a null revision
        revid_list = list(
            filter(lambda revid: not breezy.revision.is_null(revid), revid_list)
        )
        parent_map = self._branch.repository.get_graph().get_parent_map(revid_list)
        # We need to return the answer in the same order as the input,
        # less any ghosts.
        present_revids = [revid for revid in revid_list if revid in parent_map]
        rev_list = self._branch.repository.get_revisions(present_revids)

        return [self._change_from_revision(rev) for rev in rev_list]

    def _change_from_revision(self, revision):
        """
        Given a breezy Revision, return a processed "change" for use in
        templates.
        """
        message, short_message = clean_message(revision.message)

        if self._branch_tags is None:
            self._branch_tags = self._branch.tags.get_reverse_tag_dict()

        revtags = None
        if revision.revision_id in self._branch_tags:
            # tag.sort_* functions expect (tag, data) pairs, so we generate them,
            # and then strip them
            tags = [(t, None) for t in self._branch_tags[revision.revision_id]]
            sort_func = getattr(tag, "sort_natural", None)
            if sort_func is None:
                tags.sort()
            else:
                sort_func(self._branch, tags)
            revtags = ", ".join([t[0] for t in tags])

        entry = {
            "revid": revision.revision_id,
            "date": datetime.datetime.fromtimestamp(revision.timestamp),
            "utc_date": datetime.datetime.utcfromtimestamp(revision.timestamp),
            "timestamp": revision.timestamp,
            "committer": revision.committer,
            "authors": revision.get_apparent_authors(),
            "branch_nick": revision.properties.get("branch-nick", None),
            "short_comment": short_message,
            "comment": revision.message,
            "comment_clean": [util.html_clean(s) for s in message],
            "parents": revision.parent_ids,
            "bugs": [
                bug.split()[0]
                for bug in revision.properties.get("bugs", "").splitlines()
            ],
            "tags": revtags,
        }
        if isinstance(revision, breezy.foreign.ForeignRevision):
            foreign_revid, mapping = (revision.foreign_revid, revision.mapping)
        elif b":" in revision.revision_id:
            try:
                foreign_revid, mapping = (
                    breezy.foreign.foreign_vcs_registry.parse_revision_id(
                        revision.revision_id
                    )
                )
            except breezy.errors.InvalidRevisionId:
                foreign_revid = None
                mapping = None
        else:
            foreign_revid = None
        if foreign_revid is not None:
            entry["foreign_vcs"] = mapping.vcs.abbreviation
            entry["foreign_revid"] = mapping.vcs.show_foreign_revid(foreign_revid)
        return util.Container(entry)

    def get_file_changes(self, entry):
        if entry.parents:
            old_revid = entry.parents[0].revid
        else:
            old_revid = breezy.revision.NULL_REVISION
        return self.file_changes_for_revision_ids(old_revid, entry.revid)

    def add_changes(self, entry):
        changes = self.get_file_changes(entry)
        entry.changes = changes

    def get_file(self, path, revid):
        """Returns (path, filename, file contents)"""
        if not isinstance(path, str):
            raise TypeError(path)
        if not isinstance(revid, bytes):
            raise TypeError(revid)
        rev_tree = self._branch.repository.revision_tree(revid)
        display_path = path
        if not display_path.startswith("/"):
            path = "/" + path
        return (
            display_path,
            breezy.osutils.basename(path),
            rev_tree.get_file_text(path),
        )

    def file_changes_for_revision_ids(self, old_revid, new_revid):
        """
        Return a nested data structure containing the changes in a delta::

            added: list((filename)),
            renamed: list((old_filename, new_filename)),
            deleted: list((filename)),
            modified: list((filename)),
            text_changes: list((filename)),
        """
        repo = self._branch.repository
        if (
            breezy.revision.is_null(old_revid)
            or breezy.revision.is_null(new_revid)
            or old_revid == new_revid
        ):
            old_tree, new_tree = map(repo.revision_tree, [old_revid, new_revid])
        else:
            old_tree, new_tree = repo.revision_trees([old_revid, new_revid])

        reporter = FileChangeReporter(old_tree, new_tree)

        breezy.delta.report_changes(new_tree.iter_changes(old_tree), reporter)

        return util.Container(
            added=sorted(reporter.added, key=lambda x: x.filename),
            renamed=sorted(reporter.renamed, key=lambda x: x.new_filename),
            removed=sorted(reporter.removed, key=lambda x: x.filename),
            modified=sorted(reporter.modified, key=lambda x: x.filename),
            text_changes=sorted(reporter.text_changes, key=lambda x: x.filename),
        )



# Copyright (C) 2008  Canonical Ltd.
#                     (Authored by Martin Albisetti <argentina@gmail.com>)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

from .. import search
from ..controllers import TemplatedBranchView


class SearchUI(TemplatedBranchView):
    """

    Class to output progressive search result terms.
    """

    template_name = "search"

    def get_values(self, path, kwargs, response):
        """
        Default method called from the search box as /search URL

        Returns a list of suggested search terms parsed through the
        templating engine.
        """
        terms = []
        query = kwargs["query"]
        if len(query) > 0:
            terms = search.search_revisions(self._branch.branch, query, True)
            if terms is not None:
                terms = [term[0] for term in terms]
            else:
                # Should show a 'search is not available' etc box.
                terms = []

        return {"terms": terms}



# Copyright 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA

"""Code to do some load testing against a loggerhead instance.

This is basically meant to take a list of actions to take, and run it against a
real host, and see how the results respond.::

    {"parameters": {
         "base_url": "http://localhost:8080",
     },
     "requests": [
        {"thread": "1", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"}
     ],
    }

All threads have a Queue length of 1. When a third request for a given thread
is seen, no more requests are queued until that thread finishes its current
job. So this results in all requests being issued sequentially::

        {"thread": "1", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"}

While this would cause all requests to be sent in parallel:

        {"thread": "1", "relpath": "/changes"},
        {"thread": "2", "relpath": "/changes"},
        {"thread": "3", "relpath": "/changes"},
        {"thread": "4", "relpath": "/changes"}

This should keep 2 threads pipelined with activity, as long as they finish in
approximately the same speed. We'll start the first thread running, and the
second thread, and queue up both with a second request once the first finishes.
When we get to the third request for thread "1", we block on queuing up more
work until the first thread 1 request has finished.
        {"thread": "1", "relpath": "/changes"},
        {"thread": "2", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"},
        {"thread": "2", "relpath": "/changes"},
        {"thread": "1", "relpath": "/changes"},
        {"thread": "2", "relpath": "/changes"}

There is not currently a way to say "run all these requests keeping exactly 2
threads active". Though if you know the load pattern, you could approximate
this.
"""

import json
import threading
import time
from queue import Empty, Queue

from breezy import errors, transport, urlutils

try:
    from breezy.transport import NoSuchFile
except ImportError:
    from breezy.errors import NoSuchFile

# This code will be doing multi-threaded requests against breezy.transport
# code. We want to make sure to load everything ahead of time, so we don't get
# lazy-import failures
_ = transport.get_transport("http://example.com")


class RequestDescription(object):
    """Describes info about a request."""

    def __init__(self, descrip_dict):
        self.thread = descrip_dict.get("thread", "1")
        self.relpath = descrip_dict["relpath"]


class RequestWorker(object):
    """Process requests in a worker thread."""

    _timer = time.time

    def __init__(self, identifier, blocking_time=1.0, _queue_size=1):
        self.identifier = identifier
        self.queue = Queue(_queue_size)
        self.start_time = self.end_time = None
        self.stats = []
        self.blocking_time = blocking_time

    def step_next(self):
        url = self.queue.get(True, self.blocking_time)
        if url == "<noop>":
            # This is usually an indicator that we want to stop, so just skip
            # this one.
            self.queue.task_done()
            return
        self.start_time = self._timer()
        success = self.process(url)
        self.end_time = self._timer()
        self.update_stats(url, success)
        self.queue.task_done()

    def run(self, stop_event):
        while not stop_event.isSet():
            try:
                self.step_next()
            except Empty:
                pass

    def process(self, url):
        base, path = urlutils.split(url)
        t = transport.get_transport(base)
        try:
            # TODO: We should probably look into using some part of
            #       blocking_timeout to decide when to stop trying to read
            #       content
            t.get_bytes(path)
        except (errors.TransportError, NoSuchFile):
            return False
        return True

    def update_stats(self, url, success):
        self.stats.append((url, success, self.end_time - self.start_time))


class ActionScript(object):
    """This tracks the actions that we want to perform."""

    _worker_class = RequestWorker
    _default_base_url = "http://localhost:8080"
    _default_blocking_timeout = 60.0

    def __init__(self):
        self.base_url = self._default_base_url
        self.blocking_timeout = self._default_blocking_timeout
        self._requests = []
        self._threads = {}
        self.stop_event = threading.Event()

    @classmethod
    def parse(cls, content):
        script = cls()
        if isinstance(content, bytes):
            content = content.decode("UTF-8")
        json_dict = json.loads(content)
        if "parameters" not in json_dict:
            raise ValueError('Missing "parameters" section')
        if "requests" not in json_dict:
            raise ValueError('Missing "requests" section')
        param_dict = json_dict["parameters"]
        request_list = json_dict["requests"]
        base_url = param_dict.get("base_url", None)
        if base_url is not None:
            script.base_url = base_url
        blocking_timeout = param_dict.get("blocking_timeout", None)
        if blocking_timeout is not None:
            script.blocking_timeout = blocking_timeout
        for request_dict in request_list:
            script.add_request(request_dict)
        return script

    def add_request(self, request_dict):
        request = RequestDescription(request_dict)
        self._requests.append(request)

    def _get_worker(self, thread_id):
        if thread_id in self._threads:
            return self._threads[thread_id][0]
        handler = self._worker_class(
            thread_id, blocking_time=self.blocking_timeout / 10.0
        )

        t = threading.Thread(
            target=handler.run, args=(self.stop_event,), name="Thread-%s" % (thread_id,)
        )
        self._threads[thread_id] = (handler, t)
        t.start()
        return handler

    def finish_queues(self):
        """Wait for all queues of all children to finish."""
        for h, t in self._threads.values():
            h.queue.join()

    def stop_and_join(self):
        """Stop all running workers, and return.

        This will stop even if workers still have work items.
        """
        self.stop_event.set()
        for h, t in self._threads.values():
            # Signal the queue that it should stop blocking, we don't have to
            # wait for the queue to empty, because we may see stop_event before
            # we see the <noop>
            h.queue.put("<noop>")
            # And join the controlling thread
            for i in range(10):
                t.join(self.blocking_timeout / 10.0)
                if not t.is_alive():
                    break

    def _full_url(self, relpath):
        return self.base_url + relpath

    def run(self):
        self.stop_event.clear()
        for request in self._requests:
            full_url = self._full_url(request.relpath)
            worker = self._get_worker(request.thread)
            worker.queue.put(full_url, True, self.blocking_timeout)
        self.finish_queues()
        self.stop_and_join()


def run_script(filename):
    with open(filename, "rb") as f:
        content = f.read()
    script = ActionScript.parse(content)
    script.run()
    return script



# Copyright 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from breezy import tests

from ..util import html_escape, html_format


class TestHTMLEscaping(tests.TestCase):
    def test_html_escape(self):
        self.assertEqual("foo &quot;&#39;&lt;&gt;&amp;", html_escape("foo \"'<>&"))

    def test_html_format(self):
        self.assertEqual(
            '<foo bar="baz&quot;&#39;">&lt;baz&gt;&amp;</foo>',
            html_format('<foo bar="%s">%s</foo>', "baz\"'", "<baz>&"),
        )



# Copyright 2022 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from breezy import tests

from ..highlight import highlight


class TestHighLight(tests.TestCase):
    def test_no_highlighting_for_big_texts(self):
        rv = highlight(
            path="",
            text="text\n" * 102401,  # bigger than MAX_HIGHLIGHT_SIZE
            encoding="utf-8",
        )
        self.assertIsInstance(rv, list)
        self.assertLength(102401, rv)
        # no highlighting applied
        for item in rv:
            self.assertEqual("text\n", item)



#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#
"""Support for Zope Page Templates using the Chameleon library."""

import os
import re

from chameleon import PageTemplate
from importlib.resources import files

_zpt_cache: dict[str, "TemplateWrapper"] = {}


def zpt(tfile):
    tinstance = _zpt_cache.get(tfile)
    stat = os.stat(tfile)
    if tinstance is None or tinstance.stat != stat:
        with open(tfile) as tf:
            text = tf.read()
        text = re.sub(r"\s*\n\s*", "\n", text)
        text = re.sub(r"[ \t]+", " ", text)
        tinstance = _zpt_cache[tfile] = TemplateWrapper(PageTemplate(text), tfile, stat)
    return tinstance


class TemplateWrapper(object):
    def __init__(self, template, filename, stat):
        self.template = template
        self.filename = filename
        self.stat = stat

    def expand(self, **info):
        return self.template(**info)

    def expand_into(self, f, **info):
        f.write(self.template(**info).encode("UTF-8"))

    @property
    def macros(self):
        return self.template.macros


def load_template(classname):
    """Searches for a template along the Python path.

    Template files must end in ".pt" and be in legitimate packages.
    Templates are automatically checked for changes and reloaded as
    necessary.
    """
    divider = classname.rfind(".")
    if divider > -1:
        package = classname[0:divider]
        basename = classname[divider + 1 :]
    else:
        raise ValueError("All templates must be in a package")

    tfile = str(files(package) / f"{basename}.pt")
    return zpt(tfile)



#
# Copyright (C) 2008  Canonical Ltd.
#                     (Authored by Martin Albisetti <argentina@gmail.com>)
# Copyright (C) 2008  Robert Collins
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

_mod_index = None


def import_search():
    global errors, _mod_index, FileTextHit, RevisionHit
    if _mod_index is not None:
        return
    try:
        from breezy.plugins.search import errors
        from breezy.plugins.search import index as _mod_index
        from breezy.plugins.search.index import FileTextHit, RevisionHit
    except ImportError:
        _mod_index = None


def search_revisions(branch, query_list, suggest=False):
    """
    Search using bzr-search plugin to find revisions matching the query.
    This can either suggest query terms, or revision ids.

    param branch: branch object to search in
    param query_list: string to search
    param suggest: Optional flag to request suggestions instead of results
    return: A list for results, either revision ids or terms
    """
    import_search()
    if _mod_index is None:
        return None  # None indicates could-not-search
    try:
        index = _mod_index.open_index_branch(branch)
    except errors.NoSearchIndex:
        return None  # None indicates could-not-search
    query = query_list.split(" ")
    query = [(term,) for term in query]
    revid_list = []
    index._branch.lock_read()

    try:
        if suggest:
            terms = index.suggest(query)
            terms = list(terms)
            terms.sort()
            return terms
        else:
            for result in index.search(query):
                if isinstance(result, FileTextHit):
                    revid_list.append(result.text_key[1])
                elif isinstance(result, RevisionHit):
                    revid_list.append(result.revision_key[0])
            return list(set(revid_list))
    finally:
        index._branch.unlock()



import os

from .test_simple import BasicTests


class TestCornerCases(BasicTests):
    """Tests that exercise various corner cases."""

    def addFileAndCommit(self, filename, commit_msg):
        """Make a trivial commit that has 'msg' as its commit message.

        The commit adds a file called 'myfilename' containing the string
        'foo'.
        """
        self.build_tree_contents([(filename, "foo")])
        self.tree.add(filename)
        self.tree.commit(message=commit_msg)

    def test_revision_only_changing_execute_bit(self):
        """Check that a commit that only changes the execute bit of a file
        does not break the rendering."""
        self.createBranch()

        # Just a commit to have a file to change the execute bit of.
        msg = "a very exciting commit message"
        self.addFileAndCommit("myfilename", msg)

        # Make a commit that changes the execute bit of 'myfilename'.
        os.chmod("myfilename", 0o755)
        newrevid = self.tree.commit(message="make something executable")

        # Check that it didn't break things.
        app = self.setUpLoggerhead()
        res = app.get("/revision/" + newrevid.decode("utf-8"))
        res.mustcontain(b"executable")

    def test_revision_escapes_commit_message(self):
        """XXX."""
        self.createBranch()

        msg = b"<b>hi</b>"
        self.addFileAndCommit("myfilename", msg)
        app = self.setUpLoggerhead()
        res = app.get("/revision/1")
        self.assertNotIn(msg, res.body)

    def test_empty_commit_message(self):
        """Check that an empty commit message does not break the rendering."""
        self.createBranch()

        # Make a commit that has an empty message.
        self.addFileAndCommit("myfilename", "")

        # Check that it didn't break things.
        app = self.setUpLoggerhead()
        res = app.get("/changes")
        # It's not much of an assertion, but we only really care about
        # "assert not crashed".
        res.mustcontain("1")

    def test_whitespace_only_commit_message(self):
        """Check that a whitespace-only commit message does not break the
        rendering."""
        self.createBranch()

        # Make a commit that has a whitespace only message.
        self.addFileAndCommit("myfilename", "   ")

        # Check that it didn't break things.
        app = self.setUpLoggerhead()
        res = app.get("/changes")
        # It's not much of an assertion, but we only really care about
        # "assert not crashed".
        res.mustcontain("1")



# Copyright (C) 2008  Guillermo Gonzalez <guillo.gonzo@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

from ..controllers.error_ui import ErrorUI


class ErrorHandlerApp(object):
    """Class for WSGI error logging middleware."""

    msg = " %s.%s: %s \n"

    def __init__(self, application, **kwargs):
        self.application = application

    def __call__(self, environ, start_response):
        try:
            return self.application(environ, start_response)
        except BaseException:
            # test if exc_info has been set, in the case that
            # the error is caused before BranchWSGGIApp middleware
            if "exc_info" in environ.keys() and "branch" in environ.keys():
                # Log and/or report any application errors
                return self.handle_error(environ, start_response)
            else:
                # simply propagate the error, this is logged
                # by paste.httpexceptions.TransLogger middleware
                raise

    def handle_error(self, environ, start_response):
        """Exception handler."""
        self.log_error(environ)
        return errapp(environ, start_response)

    def log_error(self, environ):
        exc_type, exc_object, exc_tb = environ["exc_info"]
        logger = environ["branch"].log
        logger.exception(self.msg, exc_type.__module__, exc_type.__name__, exc_object)


def errapp(environ, start_response):
    """Default (and trivial) error handling WSGI application."""
    c = ErrorUI(environ["branch"], environ["exc_info"])
    return c(environ, start_response)



#
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import logging
import mimetypes

from breezy.errors import NoSuchRevision

try:
    from breezy.transport import NoSuchFile
except ImportError:
    from breezy.errors import NoSuchFile

from breezy import urlutils
from paste import httpexceptions
from paste.request import path_info_pop

from ..controllers import TemplatedBranchView

log = logging.getLogger("loggerhead.controllers")


class DownloadUI(TemplatedBranchView):
    def encode_filename(self, filename):
        return urlutils.escape(filename)

    def get_args(self, environ):
        args = []
        while True:
            arg = path_info_pop(environ)
            if arg is None:
                break
            args.append(arg)
        return args

    def __call__(self, environ, start_response):
        # /download/<rev_id>/<filename>
        h = self._history
        args = self.get_args(environ)
        if len(args) < 2:
            raise httpexceptions.HTTPMovedPermanently(
                self._branch.absolute_url("/changes")
            )
        revid = h.fix_revid(args[0])
        try:
            path, filename, content = h.get_file("/".join(args[1:]), revid)
        except (NoSuchFile, NoSuchRevision):
            raise httpexceptions.HTTPNotFound()
        mime_type, encoding = mimetypes.guess_type(filename)
        if mime_type is None:
            mime_type = "application/octet-stream"
        self.log.info(
            "/download %s @ %s (%d bytes)", path, h.get_revno(revid), len(content)
        )
        encoded_filename = self.encode_filename(filename)
        headers = [
            ("Content-Type", mime_type),
            ("Content-Length", str(len(content))),
            (
                "Content-Disposition",
                "attachment; filename*=utf-8''%s" % (encoded_filename,),
            ),
        ]
        start_response("200 OK", headers)
        return [content]


class DownloadTarballUI(DownloadUI):
    def __call__(self, environ, start_response):
        """Stream a tarball from a bazaar branch."""
        # Tried to re-use code from downloadui, not very successful
        if not self._branch.export_tarballs:
            raise httpexceptions.HTTPForbidden("Tarball downloads are not allowed")
        archive_format = "tgz"
        history = self._history
        self.args = self.get_args(environ)
        if len(self.args):
            revid = history.fix_revid(self.args[0])
            version_part = "-r" + self.args[0]
        else:
            revid = self.get_revid()
            version_part = ""
        # XXX: Perhaps some better suggestion based on the URL or path?
        #
        # TODO: Perhaps set the tarball suggested mtime to the revision
        # mtime.
        root = self._branch.friendly_name or "branch"
        filename = root + version_part + "." + archive_format
        encoded_filename = self.encode_filename(filename)
        headers = [
            ("Content-Type", "application/octet-stream"),
            (
                "Content-Disposition",
                "attachment; filename*=utf-8''%s" % (encoded_filename,),
            ),
        ]
        start_response("200 OK", headers)
        tree = history._branch.repository.revision_tree(revid)
        return tree.archive(root=root, format=archive_format, name=filename)



#
# Copyright (C) 2008, 2009 Canonical Ltd.
# Copyright (C) 2006  Robey Pointer <robey@lag.net>
# Copyright (C) 2006  Goffredo Baroncelli <kreijack@inwind.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

import json

from breezy import urlutils
from paste.httpexceptions import HTTPServerError

from .. import util
from ..controllers import TemplatedBranchView


class ChangeLogUI(TemplatedBranchView):
    template_name = "changelog"

    def get_values(self, path, kwargs, headers):
        history = self._history
        revid = self.get_revid()
        filter_path = kwargs.get("filter_path", path)
        query = kwargs.get("q", None)
        start_revid = history.fix_revid(kwargs.get("start_revid", None))
        orig_start_revid = start_revid
        pagesize = 20  # int(config.get('pagesize', '20'))
        search_failed = False

        try:
            revid, start_revid, revid_list = history.get_view(
                revid, start_revid, filter_path, query, extra_rev_count=pagesize + 1
            )
            util.set_context(kwargs)

            if (query is not None) and (len(revid_list) == 0):
                search_failed = True

            if len(revid_list) == 0:
                scan_list = revid_list
            else:
                if revid in revid_list:  # XXX is this always true?
                    i = revid_list.index(revid)
                else:
                    i = None
                scan_list = revid_list[i:]
            change_list = scan_list[:pagesize]
            changes = list(history.get_changes(change_list))
            data = {}
            for i, c in enumerate(changes):
                c.index = i
                data[str(i)] = urlutils.quote(
                    urlutils.quote_from_bytes(c.revid, safe="")
                )
        except BaseException as e:
            self.log.exception("Exception fetching changes")
            raise HTTPServerError("Could not fetch changes") from e

        navigation = util.Container(
            pagesize=pagesize,
            revid=revid,
            start_revid=start_revid,
            revid_list=revid_list,
            filter_path=filter_path,
            scan_url="/changes",
            branch=self._branch,
            feed=True,
            history=history,
        )
        if query is not None:
            navigation.query = query
        util.fill_in_navigation(navigation)

        # Directory Breadcrumbs
        directory_breadcrumbs = util.directory_breadcrumbs(
            self._branch.friendly_name, self._branch.is_root, "changes"
        )

        show_tag_col = False
        for change in changes:
            if change.tags is not None:
                show_tag_col = True
                break

        return {
            "branch": self._branch,
            "changes": changes,
            "show_tag_col": show_tag_col,
            "data": json.dumps(data),
            "util": util,
            "history": history,
            "revid": revid,
            "navigation": navigation,
            "filter_path": filter_path,
            "start_revid": start_revid,
            "viewing_from": (orig_start_revid is not None)
            and (orig_start_revid != history.last_revid),
            "query": query,
            "search_failed": search_failed,
            "url": self._branch.context_url,
            "directory_breadcrumbs": directory_breadcrumbs,
        }



#
# Copyright (C) 2009  Peter Bui <pnutzh4x0r@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335  USA
#

from html import escape

import breezy.osutils
from pygments import highlight as _highlight_func
from pygments.formatters import HtmlFormatter
from pygments.lexers import TextLexer, guess_lexer, guess_lexer_for_filename
from pygments.util import ClassNotFound

DEFAULT_PYGMENT_STYLE = "colorful"

# Trying to highlight very large files using pygments was killing
# loggerhead on launchpad.net, because pygments isn't very fast.
# So we only highlight files if they're 512K or smaller.
MAX_HIGHLIGHT_SIZE = 512000


def highlight(path, text, encoding, style=DEFAULT_PYGMENT_STYLE):
    """
    Returns a list of highlighted (i.e. HTML formatted) strings.
    """

    if len(text) > MAX_HIGHLIGHT_SIZE:
        return list(map(escape, breezy.osutils.split_lines(text)))

    formatter = HtmlFormatter(style=style, nowrap=True, classprefix="pyg-")

    try:
        lexer = guess_lexer_for_filename(path, text[:1024], encoding=encoding)
    except (ClassNotFound, ValueError):
        try:
            lexer = guess_lexer(text[:1024], encoding=encoding)
        except (ClassNotFound, ValueError):
            lexer = TextLexer(encoding=encoding)

    hl_lines = _highlight_func(text, lexer, formatter)
    hl_lines = breezy.osutils.split_lines(hl_lines)

    return hl_lines
